# noqa: V841
# noqa: V841i
# noqa: V102
# noqa: V103
# noqa: V101
# noqa: V105
import dataclasses
import enum

from p2g import axis
from p2g import coords
from p2g import err
from p2g import nd
from p2g import stat


class MovementSpace(enum.IntEnum):
    UNDEFINED = enum.auto()
    RELATIVE = enum.auto()
    WORK = enum.auto()
    MACHINE = enum.auto()
    R9810 = enum.auto()

    def __str__(self):
        return self.name.lower()


class MovementOrder(enum.IntEnum):
    UNDEFINED = enum.auto()
    XYZ = enum.auto()
    Z_LAST = enum.auto()
    Z_FIRST = enum.auto()

    def __str__(self):
        return self.name.lower()


@dataclasses.dataclass(eq=True, frozen=True)
class GotoWorker:
    _space: MovementSpace
    _order: MovementOrder
    _feed: float
    _probe: bool
    _mcode: str
    _delay: int

    def accumulate_coords(self, cos):
        for aname, value in cos:
            yield f"{aname}{value.to_gcode(nd.NodeModifier.ARGUMENT)}"

    def accumulate_probe(self):
        if self._probe:
            yield "G31 G55"
        else:
            yield "G01 G55"

    def accumulate_mcode(self):
        if self._mcode:
            yield self._mcode

    def accumulate_feed(self):
        if self._feed == 0.0:
            err.compiler("Need feed rate.")
        yield f"F{nd.to_gcode(self._feed)}"

    def update(self, key, value):
        xxx = {**self.__dict__, key: value}
        return GotoWorker(**xxx)

    def _update_space(self, newv):
        if self._space not in (MovementSpace.UNDEFINED, newv):
            err.compiler(f"Conflicting spaces for goto, '{self._space}' and '{newv}'.")
        return self.update("_space", newv)

    def _update_order(self, newv):
        if self._order not in (MovementOrder.UNDEFINED, newv):
            err.compiler(f"Conflicting orders for goto, '{self._order}' and '{newv}'.")
        return self.update("_order", newv)

    @property  # noqa
    def relative(self):
        return self._update_space(MovementSpace.RELATIVE)

    @property  # noqa
    def work(self):
        return self._update_space(MovementSpace.WORK)  # noqa: V841

    @property  # noqa
    def r9810(self):
        return self._update_space(MovementSpace.R9810)

    @property  # noqa
    def machine(self):
        return self._update_space(MovementSpace.MACHINE)

    @property  # noqa
    def z_last(self):
        return self._update_order(MovementOrder.Z_LAST)

    @property  # noqa
    def z_first(self):
        return self._update_order(MovementOrder.Z_FIRST)

    @property  # noqa
    def all(self):
        return self._update_order(MovementOrder.XYZ)

    @property  # noqa
    def xyz(self):
        return self._update_order(MovementOrder.XYZ)

    @property  # noqa
    def probe(self):
        return self.update("_probe", True)

    def feed(self, _feed):  # noqa
        return self.update("_feed", _feed)

    def mcode(self, m_code):  # noqa
        return self.update("_mcode", str(m_code))

    def delay(self, delay):  # noqa
        return self.update("_delay", delay)

    def accumulate_parts1(self):
        match self._space:
            case MovementSpace.RELATIVE:
                yield "G91"
                yield from self.accumulate_probe()
                yield from self.accumulate_mcode()
            case MovementSpace.MACHINE:
                yield "G90"
                yield "G53"
                yield from self.accumulate_probe()
                yield from self.accumulate_mcode()
            case MovementSpace.WORK:
                yield "G90"
                yield from self.accumulate_probe()
                yield from self.accumulate_mcode()
            case MovementSpace.R9810:
                yield "G65 R9810"
                if self._probe:
                    err.compiler("Probe with 9810 move is illegal.")
                if self._mcode:
                    err.compiler("MCODE with 9810 move is illegal.")
            case _:
                err.compiler(
                    "'goto' needs one of 'relative','machine', 'work' or 'R9810'."
                )
        yield from self.accumulate_feed()

    def do_goto_worker(self, fter, args, kwargs):
        # split out arguments we understand from
        # ones for coordinates.

        def get_coords():
            values = coords.unpack(args, kwargs)

            for aname, value in zip(axis.low_names_v(), values):
                # skip non mentioned coords
                if fter != "*" and aname not in fter:
                    continue
                # skip non there coords
                if value.is_none_constant:
                    continue
                yield aname, value

        cos = list(get_coords())
        if cos:
            stat.code(
                self.accumulate_parts1(),
                self.accumulate_coords(cos),
            )
        if self._delay:
            stat.code(f"G103 P{self._delay}")

    def __call__(self, *args, **kwargs):
        match self._order:
            case MovementOrder.UNDEFINED:
                err.compiler("'goto' needs movement order, 'all', 'z_first' or 'z_last'.")
            case MovementOrder.XYZ:
                self.do_goto_worker("*", args, kwargs)
            case MovementOrder.Z_FIRST:
                self.do_goto_worker("z", args, kwargs)
                self.do_goto_worker("xyabc", args, kwargs)
            case MovementOrder.Z_LAST:
                self.do_goto_worker("xyabc", args, kwargs)
                self.do_goto_worker("z", args, kwargs)

    def __repr__(self):
        return (
            f"goto space={self._space} order={self._order.value} "
            f"_feed={self._feed} probe={self._probe} mcode='{self._mcode}'"
        )


goto = GotoWorker(
    _space=MovementSpace.UNDEFINED,
    _order=MovementOrder.UNDEFINED,
    _feed=0,
    _delay=0,
    _probe=False,
    _mcode="",
)



# turn fstrings for print into the right runes for dprnt
import ast
import re

import p2g

from p2g import axis
from p2g import gbl
from p2g import nd
from p2g import op


# takes a python format and turns it into something
# in the new formatty way.
#    ###.## makes f3.2
#     ?xyz  makes xyz only show up between elements.
#     A     makes axis letter print
#     !     makes axis number print
#   anything else is just copied out.
def py_fmt_to_gcode(fmt):

    if fmt in ("", "f"):
        return "#####.###"

    if fmt[0].isdigit() or fmt[0] == ".":
        # standard f fmt.

        groups = re.match("(?P<n_total>\\d*).(?P<n_frac>\\d*)(f?)(?P<rest>:.*)?", fmt)
        if groups:
            gdict = groups.groupdict()
            n_total = int(gdict["n_total"])
            n_frac = gdict.get("n_frac")

            if n_frac == "":
                n_frac = 0
            else:
                n_frac = int(gdict["n_frac"])
            restp = gdict["rest"] if gdict["rest"] else ""
            n_mant = max(n_total - n_frac - 1, 1)
            return ("#" * n_mant) + "." + ("#" * n_frac) + restp
    return fmt


@gbl.g2s
def expand_magic_chars(gcode_fmt, vector_idx):
    for el in gcode_fmt:
        if el == "@":
            yield axis.hi_names(vector_idx)
        elif el == "!":
            yield str(vector_idx)
        else:
            yield el


def get_number_insert(val, prefix_digits, suffix_digits):
    konst = op.get_constant(val)

    if konst:
        pyfmt = f"{prefix_digits+suffix_digits + 1}.{suffix_digits}f"
        number_insert = format(konst.to_float(), pyfmt)
    else:
        node = op.reload(val)
        number_insert = nd.to_gcode(node) + f"[{prefix_digits}{suffix_digits}]"
    return number_insert


# passed a magical my format gcode format,
# like "####.###, " which will generate
# a dprint foo[42], bar[42], zap[42]
# etc
@gbl.g2s
def format_for_one(val, gcode_fmt, idx):

    groups = re.match(
        "(?P<prefix>[^#?]*)"
        "((?P<mant>#*)\\.?(?P<frac>#*))?"
        "(?P<suffix>[^?]*)(\\?(?P<between>.*))?",
        gcode_fmt,
    )

    if groups:
        gdict = groups.groupdict()
        if idx > 0 and gdict["between"]:
            yield from expand_magic_chars(gdict["between"], idx)
        yield from (expand_magic_chars(gdict["prefix"], idx))
        yield from (get_number_insert(val, len(gdict["mant"]), len(gdict["frac"])))
        yield from (expand_magic_chars(gdict["suffix"], idx))


@gbl.g2s
def format_expand(val, gcode_fmt):
    for (idx, el) in enumerate(val.everything()):
        yield format_for_one(el, gcode_fmt, idx)


def format_for_n(val, fmt):
    if isinstance(val, str):
        return val
    if not fmt:
        fmt = "7.2f"
    gcode_fmt = py_fmt_to_gcode(fmt)
    if op.get_nelements(val) > 1:
        return format_expand(val, gcode_fmt)
    return format_for_one(p2g.scalar.wrap_scalar(val), gcode_fmt, 0)


def one_float(interp, value):

    val = interp.visit(value.value)

    fmt = interp.visit(value.format_spec)
    el = format_for_n(val, fmt)
    return el


@gbl.g2s
def joinedstr(interp, node):
    for value in node.values:
        match type(value):
            case ast.Constant:
                yield value.value
            case _:
                el = one_float(interp, value)
                yield el



import p2g

# stop with alarm code showing largest
# flute count in table.


def maxflutes():

    mx_flutes = p2g.Var(p2g.haas.TOOL_TBL_FLUTES[0])
    for n_flutes in p2g.haas.TOOL_TBL_FLUTES:
        if n_flutes > mx_flutes:
            mx_flutes = n_flutes

    p2g.haas.MESSAGE.var = mx_flutes



from typing import Optional
import p2g


import usrlib as usr
from p2g import sys
from p2g.sys import print


class SearchConstraint:
    # position roughly center above start of search.
    above: p2g.Vec
    # min size of thing to look for.

    amin: p2g.Vec
    # max size to search in one dimension for the thing
    # to probe.  p2g.RVALnstains max size of things measured.
    amax: p2g.Vec

    # when measuring z, how far to
    # move from an xy edge inwards to work it out.
    indent: p2g.Vec

    # steps between probes

    delta: p2g.Vec
    # safe distance above probed z surface to move
    skim: p2g.Vec

    # how far to backoff before probing slowly
    backoff: p2g.Vec

    # how far to look down for a surface once we've
    # already located the rough z0
    search_depth: p2g.Vec
    # how far to look down and not found is a not there
    #
    found_if_below: p2g.Vec

    def __init__(
        self,
        amin: p2g.Vec,
        amax: p2g.Vec,
        above: p2g.Vec,
        delta: p2g.Vec,
        skim=None,
        indent: Optional[p2g.Vec] = None,
        backoff: Optional[p2g.Vec] = None,
        search_depth=None,
        zslack=None,
        found_if_below=None,
    ):

        if zslack is None:
            zslack = p2g.Const(-1.0)

        if search_depth is None:
            search_depth = p2g.Const(-0.1)

        if found_if_below is None:
            found_if_below = search_depth * 0.8

        if backoff is None:
            backoff = p2g.Const(x=0.05, y=0.05, z=0.05)

        if indent is None:
            indent = round(amax * 0.1, 1)

        if skim is None:
            skim = p2g.Const(0.15)

        self.delta = delta

        self.amin = amin
        self.indent = indent
        self.amax = amax
        self.above = above
        self.skim = skim
        self.backoff = backoff
        self.search_depth = search_depth
        self.zslack = zslack
        self.found_if_below = found_if_below
        self.edge_search_depth = -0.1

    def comment(self):
        return [
            "Search Constraints:",
            "start:",
            f"  {self.above}",
            "boundary:",
            f"  x= (-{(self.amax[0] * 0.5)}..-{(self.amin[0] * 0.5)})"
            f"..({(self.amin[0] * 0.5)}..{(self.amax[0] * 0.5)})",
            f"  y= (-{(self.amax[1] * 0.5)}..-{(self.amin[1] * 0.5)})"
            f"..({(self.amin[1] * 0.5)}..{(self.amax[1] * 0.5)})",
            f"  z= ({(-self.amax[2])}..{(-self.amin[2])})",
            "indent:",
            f"  {self.indent}",
            "delta:",
            f"  {self.delta}",
        ]


sch = SearchConstraint(
    # minimum size expected
    amin=p2g.Const(x=8.0, y=4.0, z=1.0),
    # maximum size expected
    amax=p2g.Const(x=16.0, y=9.0, z=3.2),
    delta=p2g.Const(x=1.0, y=1.0),
    above=usr.MABS_ABOVE_VICE,
)


# describe relationship between x&y and left,right,far,near
class Dinfo:
    name: str
    opposite: str
    cur_axis: int
    dxdy: tuple[float, float]

    def __init__(self, name, opposite, cur_axis, dxdy):
        self.name = name
        self.opposite = opposite
        self.cur_axis = cur_axis
        self.dxdy = dxdy


dinfo = {
    "left": Dinfo("left", "right", 0, (-1, None)),
    "right": Dinfo("right", "left", 0, (1, None)),
    "far": Dinfo("far", "near", 1, (None, 1)),
    "near": Dinfo("near", "far", 1, (None, -1)),
}


def show3(v):
    return f"{v:###.####?, }"


def find_top(wcs):
    p2g.comment("find top z roughly.")

    usr.goto(z=0)

    p2g.comment("just above workpiece surface.")
    usr.goto_down(sch.above)

    # fast find move down to min search distance
    usr.fast_work_probe(z=sch.above.z + sch.zslack)

    p2g.com("make wcs become ~0,0,0 at tdc")
    wcs.xyz = +p2g.haas.SKIP_POS

    usr.goto(z=sch.skim)
    p2g.comment(
        "now work.xyz should be 0, with z at skim",
        "distance, physically roughly tdc",
    )
    print("first estimate:")
    print(f"wcs now       {wcs[:2]:###.####?, }")


#    {wcs.x:5.2f}{wcs.y:5.2f}{wcs.z:5.2f}")


# finds the edge asked for by st and sch using
# edge_name as direction.
# asssumes probe is skimming over
# start pos, moves in direction. finds edge.
# stops there.


def fast_find(edge_name):
    di = dinfo[edge_name]
    start_search = p2g.Const[2](sch.amin.xy * di.dxdy / 2)
    stop_search = p2g.Const[2](sch.amax.xy * di.dxdy / 2)
    delta = sch.delta * di.dxdy

    # work out how many iterations to get from
    # start position to end position.

    gap = (stop_search - start_search)[di.cur_axis]
    number_bumps = gap / delta[di.cur_axis]

    its = p2g.Var(abs(number_bumps) + 1)

    p2g.comment(
        f"Fast find {di.name} edge:",
        " Starting from min possible dimension,",
        f" move probe {di.name}wards by",
        " calculated delta. If surface not found,",
        " then done.",
    )

    usr.optional_pause(f"top of {di.name}")
    # move the interesting dimension to at least the minimum
    # distance for an edge.

    usr.goto_up(sch.amin.xy * di.dxdy * 0.5, z=sch.skim)

    while its > 0:
        usr.goto_rel(delta.xy)
        usr.fast_work_probe(z=sch.search_depth)
        if p2g.haas.SKIP_POS.z < sch.found_if_below:
            break
        usr.goto(z=sch.skim)
        its -= 1
    else:
        usr.alarm(f"search for {di.name} failed")


# backoff from the direction being inspected
# then probe fast towards the edge, backoff and
# then probe slowly.


def backoff_and_fast_then_slow_probe(edge_name, error):
    di = dinfo[edge_name]
    p2g.comment(
        f"Accurately find {di.name} edge: ",
        f" Back off a bit {di.name}wards.",
        " Move to search height.",
        f" Slowly probe towards the {di.opposite}",
        " edge.",
    )
    # away from edge.
    usr.goto_rel(sch.backoff.xy * di.dxdy)
    # back to nominal search for edge height.
    usr.goto(z=sch.search_depth)
    # fast towards edge
    usr.fast_rel_probe(xy=-sch.indent.xy * di.dxdy)
    # backoff a little
    usr.goto_rel(sch.backoff.xy * di.dxdy)
    # and slowly find edge.
    usr.slow_rel_probe(xy=-sch.indent.xy * di.dxdy)

    # no need to adjust by probe_r, since errors will
    # cancel out.
    error[di.cur_axis] += p2g.haas.SKIP_POS[di.cur_axis]
    print(f"at {edge_name.ljust(5)} error {error:###.####?, }")


def move_above_and_inwards(edge_name):
    di = dinfo[edge_name]
    p2g.comment(
        "Above surface and in:",
        f" Back off {edge_name} edge, up to skim distance.",
        " Move towards center.",
    )

    usr.goto_rel(sch.backoff.xy * di.dxdy)
    usr.goto(z=sch.skim)
    usr.goto_rel(-sch.indent.xy * di.dxdy)


def find_edge(edge_name, error):

    with sys.BSS():
        fast_find(edge_name)
        backoff_and_fast_then_slow_probe(edge_name, error)
        move_above_and_inwards(edge_name)


def find_edges(error):
    # enter at roughly 0,0,0

    find_edge("left", error)
    find_edge("near", error)
    find_edge("far", error)
    usr.goto(0, 0)
    find_edge("right", error)


def calc_center(wcs, error):
    p2g.comment(
        "The x coordinates of the left and right",
        "edge have been summed and put into",
        "error[0]. That value is double the error",
        "of the first guess, since every bit of",
        "error shifts the apparent locations of",
        "the left and right side (if the",
        "workpiece was exactly placed under the",
        "first approximation, the left and right",
        "coordinats would have been equal and",
        "opposite, so the sum would be zero. The",
        "same is true for error[1], for far and",
        "near.",
    )

    wcs.xy += error.xy / 2.0
    usr.goto(0, 0)
    p2g.comment(" final slow probe to find the surface z")
    usr.slow_work_probe(z=sch.search_depth)
    wcs.z += p2g.haas.SKIP_POS.z
    print(f"wcs done      {wcs[:3]:###.####?, }")
    usr.goto(z=1)


def vicecenter():
    p2g.Control.emacsclient = True
    p2g.Control.symbol_table = True
    p2g.comment(
        "Find center of plate in vice,",
        *sch.comment(),
    )

    with (
        sys.WCS(p2g.haas.G55) as wcs,
        usr.Probe(),
        p2g.sys.Lookahead(False),
    ):
        # start with g55 same as machine coords.
        p2g.haas.G55.xyz = (0, 0, 0)
        error = p2g.Var[2](0, 0)

        find_top(wcs)
        find_edges(error)
        calc_center(wcs, error)



# flake8: noqa
# pylint:  disable=wrong-import-order
# pylint:  disable=wrong-import-position
# isort: skip_file
# no import sorting, typeguard must do its thing before
# imports get done.
import typeguard
import p2g.main

typeguard.install_import_hook("p2g")
p2g.main.main()



from p2g import coords
from p2g import gbl
from p2g import haas
from p2g import stat
from p2g import vector


NLOOKAHEAD_OFF = 123


# pylint: disable=redefined-builtin
def print(fstr):
    stat.add_stat(stat.Dprint(fstr))


def load_tool(tool):
    stat.codenl(f"T{tool:02} M06")


class Optional:
    prev: str

    def __init__(self):
        self.prev = gbl.Control.code_prefix
        gbl.Control.code_prefix = "/"

    def __enter__(self):
        pass

    def __exit__(self, *_):
        gbl.Control.code_prefix = self.prev


class WCS:
    def __init__(self, mvar):
        self.prev = coords.Var()
        self.prev.var = haas.PREV_WCS
        self.mvar = mvar
        stat.code(f"G{(mvar.get_address() - 5221)//20 + 54}")

    def __enter__(self):
        return self.mvar

    def __exit__(self, *_):
        stat.code(f"G{self.prev.var}", comment_txt="Restore wcs")


def base_addr(new_base=None):
    #    ._wcs = (val._addr - 5221) // 20 + 54
    if new_base is not None:
        gbl.iface.ebss = new_base
    return gbl.iface.ebss


class BSS:
    quiet: bool
    prev: int

    def __init__(self, quiet=True):
        self.prev = gbl.iface.ebss
        self.quiet = quiet
        if not self.quiet:
            stat.comment(f"Save bss at {self.prev}")

    def __enter__(self):
        pass

    def __exit__(self, *_):
        gbl.iface.ebss = self.prev
        if not self.quiet:
            stat.comment(f"Restore bss to {self.prev}")


# don't fill in the comment text with the  srcline
# since we're printing it.
def message(dst, txt: str):
    stat.code(f"{dst.to_gcode()} = 101 ({txt})", comment_txt=stat.CommentGen.NONE)


def address(src):
    # special case for vecs, give address
    # anyway.
    wrapped = vector.wrap_mustbe_vector(src)
    add = wrapped.get_address()
    return add


turn_off_lookahead = [
    f"N{NLOOKAHEAD_OFF}",
    "G103 P1",
    "G04 P1",
    "G04 P1",
    "G04 P1",
    "G04 P1",
    "M99",
]


class Lookahead:
    prev_lookahead: bool
    machine_lookahead = True
    head_comment: str
    tail_comment: str
    need_off_code = False

    def __init__(self, lookahead=True):
        self.prev_lookahead = Lookahead.machine_lookahead
        self.set_lookahead(lookahead)

    @classmethod
    def on(cls):
        cls.set_lookahead(True)

    @classmethod
    def off(cls):
        cls.set_lookahead(False)

    @classmethod
    def on_exit(cls):
        cls.machine_lookahead = True

        if not cls.need_off_code:
            return
        stat.codenl(
            turn_off_lookahead,
            comment_txt=stat.CommentGen.NONE,
        )
        cls.need_off_code = False

    @classmethod
    def set_lookahead(cls, lookahead=True):
        if lookahead != Lookahead.machine_lookahead:
            if lookahead:
                stat.codenl("G103")
            else:
                cls.need_off_code = True
                stat.codenl(f"M97 P{NLOOKAHEAD_OFF}")

            Lookahead.machine_lookahead = lookahead

    def __enter__(self):
        #        self.status()
        pass

    def __exit__(self, *_):
        #        self.status()
        self.set_lookahead(self.prev_lookahead)


gbl.on_exit.append(Lookahead.on_exit)



# noqa
import abc
import typing

from p2g import axis
from p2g import err
from p2g import nd
from p2g import scalar
from p2g import stat
from p2g import symbol


class Vec(nd.EBase, nd.FakeOps):
    def __div__(self, other):
        pass  # no cover

    def __mult__(self, other):
        pass  # no cover

    def __init__(self):
        pass

    @abc.abstractmethod
    def nelements(self) -> int:
        ...

    @abc.abstractmethod
    def get_address(self) -> scalar.Scalar:
        ...

    @abc.abstractmethod
    def get_at(self, idx: scalar.Scalar) -> scalar.Scalar:
        ...

    @abc.abstractmethod
    def get_slice(self, slindex: slice) -> "Vec":
        ...

    def to_scalar(self):
        return self[0]

    def get_axes_vec(self, key):
        if key == "var":
            return TupleV(list(scalar.urange(0, self.nelements())))
        indexes = axis.name_to_indexes_list(key)
        if max(indexes) >= self.nelements():
            err.compiler("Reference to too many axes.")
        return TupleV(indexes)

    def forever(self) -> typing.Generator[scalar.Scalar, None, None]:
        while True:
            yield from self.everything()

    def __getattr__(self, key):
        if key[0] == "_":
            return object.__getattribute__(self, key)
        axes_vec = self.get_axes_vec(key)
        return sorv_from_list([self.get_at(idx) for idx in axes_vec])

    def __setattr__(self, key, value):
        if key[0] == "_":
            object.__setattr__(self, key, value)
            return
        value = wrap(value)
        axes_vec = self.get_axes_vec(key)
        for idx, src in zip(axes_vec, value.forever()):
            self.set_at(idx, src)

    def __getitem__(self, scalar_index) -> scalar.Scalar:
        if isinstance(scalar_index, slice):
            return typing.cast(scalar.Scalar, self.get_slice(scalar_index))
        return self.get_at(scalar.wrap_scalar(scalar_index))


class TupleV(Vec):
    _guts: list[scalar.Scalar]
    _from_user: bool

    def __init__(self, value, from_user=False):
        super().__init__()
        object.__setattr__(self, "_guts", [])
        object.__setattr__(self, "_from_user", from_user)
        for el in value:
            self._guts.append(scalar.wrap_scalar(el))

    def __div__(self, _other):
        return self  # no cover

    def get_address(self):
        return self._guts[0].get_address()

    def everything(self) -> typing.Generator[scalar.Scalar, None, None]:
        yield from self._guts

    def get_at(self, idx: scalar.Scalar) -> scalar.Scalar:
        return self._guts[int(idx)]

    def nelements(self) -> int:
        return len(self._guts)

    def to_symtab_entry(self, _):
        res = []
        tlen = 0
        for el in self._guts:
            nextstr = nd.to_gcode(el, nd.NodeModifier.F3X3)
            tlen += len(nextstr)
            if tlen > 21:
                res.append("...")
                break
            res.append(nextstr)
        return symbol.Group.USER, ",".join(res)

    def get_slice(self, slindex: slice):
        return TupleV(self._guts[slindex])

    def __iter__(self):
        yield from self._guts

    def __str__(self):  # no cover
        return ", ".join(map(str, self._guts))


symbol.HasToSymTabEntry.register(TupleV)
# stub overwritten by op.py to avoid cycles


def local_hashop(_addr: scalar.Scalar, _idx: scalar.Scalar) -> scalar.Scalar:
    raise NotImplementedError


class MemVec(Vec):
    _size: scalar.Scalar
    _addr: scalar.Scalar
    _step: scalar.Scalar

    def __init__(self, _addr=0, _size=0, _step=1):
        super().__init__()
        self._addr = scalar.wrap_scalar(_addr)
        self._size = scalar.wrap_scalar(_size)
        self._step = scalar.wrap_scalar(_step)

    def __div__(self, _other):
        return self  # no cover

    def get_address(self):
        return self._addr

    def to_symtab_entry(self, varrefs) -> typing.Tuple[symbol.Group, str]:
        fwidth = 7
        hit_indexes: dict[int, bool] = {}
        for el, addr in enumerate(scalar.urange(self._addr, self._addr + self._size)):
            if addr in varrefs:
                hit_indexes[el] = True

        if hit_indexes and max(hit_indexes) >= len(axis.NAMES):
            return symbol.Group.VECTOR, f"#{self._addr}[{self._size}]".rjust(fwidth)
        res = []
        for idx in scalar.urange(0, self._size):
            if idx not in hit_indexes:
                continue
            res.append(
                f"#{self._addr + idx}.{axis.low_names(idx)}".rjust(fwidth),
            )
        return symbol.Group.VECTOR, " ".join(res)

    def nelements(self) -> int:
        return int(self._size)

    def everything(self) -> typing.Generator[scalar.Scalar, None, None]:
        return (
            self.get_at(scalar.wrap_scalar(idx))
            for idx in scalar.urange(0, self._size, int(self._step))
        )

    def get_at(self, idx: scalar.Scalar):
        res = local_hashop(self._addr, idx)
        if not isinstance(idx, scalar.Constant):
            return res
        fidx = int(idx)
        if not 0 <= fidx < int(self._size):
            err.compiler(
                f"Index out of range, index={fidx} size={self._size}.",
            )
        return res

    def set_at(self, idx: scalar.Scalar, src: scalar.Scalar):
        if src.is_none_constant:
            return
        stat.append_set(self.get_at(idx), src)

    def get_slice(self, slindex: slice):
        tmp = (list(range(self.nelements())))[slindex]
        step = 0
        if len(tmp) > 1:
            step = tmp[1] - tmp[0]
        addr = self._addr + tmp[0]
        size = tmp[-1] - tmp[0] + 1
        assert addr is not None
        return MemVec(addr, size, step)

    def __setitem__(self, indexes, src):
        if isinstance(indexes, slice):
            indexes = list(range(*indexes.indices(self.nelements())))
        indexes = wrap(indexes)
        src = wrap(src)
        for idx, sel in zip(indexes.everything(), src.forever()):
            stat.append_set(self[idx], sel)

    def __repr__(self):
        return f"(array  {self._addr} {self._size})"


def wrap(thing) -> typing.Union[Vec, scalar.Scalar]:
    if isinstance(thing, (TupleV, MemVec)):
        return thing
    if isinstance(thing, (list, set, dict)):
        return TupleV(thing)
    if isinstance(thing, tuple):
        return TupleV(list(thing))
    return scalar.wrap_scalar(thing)


def wrap_mustbe_vector(src):
    if isinstance(src, Vec):
        return src

    raise err.CompilerError("Only vectors have addresses.")


symbol.HasToSymTabEntry.register(MemVec)


def sorv_from_list(thing: list) -> typing.Union[Vec, scalar.Scalar]:
    if len(thing) != 1:
        return TupleV(thing)
    return thing[0]



import collections
import datetime
import pathlib
import re
import shutil
import typing

import docopt

from p2g import err
from p2g import gbl
from p2g import walkfunc


DOC = """
p2g - Turn Python into G-Code.

Usage:
  p2g [options]  <srcfile> [<dstfile>]
  p2g help [ all | topics | maint | version | location | <topic> ]
  p2g examples <dstdir>

#   For bare p2g:
#       p2g tram-rotary.py ~/_nc_/O{countdown}tr.nc
#        Makes an output of the form ~/_nc_/O1234tr.nc
#
#       p2g --func=thisone -
#        Read from stdin, look for the 'thisone' function and write to
#        to stdout.
#

Arguments:
  <srcfile>   Source python file. [default: stdin]
  <dstfile>   Destination G-Code file. [default: stdout]
               {countdown} in file name creates a decrementing prefix
               for the output file which makes looking for the .nc in
               a crowded directory less painful - it's at the top.
               (It's the number of seconds until midnight, so clear
               the directory once a day.)
  <topic>      [ all | topics | maint | version | location | <topic> ]
        # all      Print all readme.
        # topics   List all topics.
        # maint    Print maintenance options.
        # version  Show version
        # location Show absdir of main
        # <topic>  Print from readme starting at topic.




Options:
     --job=<jobname>      Olabel for output code.
     --function=<fname>   Function to be compiled,
                           default is last one in source file.
     --narrow             Emit comments on their own line,
                           makes text fit more easily into
                           a narrow program window.
     --short-filenames    Emit just the lsb of filenames.
!
!For maintenance:
!     --break              pdb.set_trace() on error.
!     --no-id=             Don't put version in outputs.
!     --verbose=<level>    Set verbosity level [default: 0]
"""


def do_examples(outdir):
    find_examples = gbl.find_ours("vicecenter.py")
    dst_dir = outdir.resolve()
    example_dir = find_examples.parent
    example_files = example_dir.glob("[a-z]*.py")
    outdir.mkdir(exist_ok=True, parents=True)
    for src in example_files:
        dst_name = outdir / src.name
        gbl.sprint(f"Copying {src} {dst_name}")
        shutil.copy(src, dst_name)
    for job in ["vicecenter", "probecalibrate", "maxflutes"]:
        rootname = dst_dir / job
        sysargs = [
            str(rootname.with_suffix(".py")),
            str(rootname.with_suffix(".nc")),
        ]
        gbl.sprint(f"Running p2g {sysargs[0]} {sysargs[1]}")
        main(sysargs)


# chop readme.txt into bits given back by
# choice of section.
def do_doc(want):

    doc = gbl.find_ours("howto.txt").read_text()
    lines = doc.split("\n")
    by_section = collections.defaultdict(list)
    # a section header comes the text, not TOC, looks like:
    # the form <number>  <title>
    want = want.lower()
    section_name = ""
    for line in lines:
        got = re.match("^\\d+[. ]+(.*)", line)
        if got:
            section_name = got.group(1).lower()
        by_section[section_name].append(line)

    if want == "topics":
        for section in by_section:
            print(f" {section}")
        return
    doneone = False

    for section, lines in by_section.items():
        if want == "all" or want in section:
            doneone = True

            for line in lines:
                print(line)
    if not doneone:
        print(f"{want} not found, topics:")
        for section in by_section:
            print(f" {section}")


def calculate_output_file_name(provided_filename):
    # if provided_filename is empty then output is stdout.
    if provided_filename is None:
        return "-"

    now = datetime.datetime.now()
    prev_midnight_secs = now.replace(hour=0, minute=0, second=0, microsecond=0)
    mins_since = (now - prev_midnight_secs).seconds // 60
    mins_togo = 24 * 60 - mins_since
    return provided_filename.replace("{countdown}", f"{mins_togo:04d}")


def do_gen(src_name, job_name, func_name, output_name):

    if func_name is None:
        func_name = "<last>"
    job_name = job_name if job_name else "O0001"
    output_name = calculate_output_file_name(output_name)
    src_path = pathlib.Path(src_name)
    gbl.v1print(f"src: {src_path}")
    gbl.v1print(f"fnc: {func_name}")
    gbl.v1print(f"job: {job_name}")
    gbl.v1print(f"out: {output_name}")

    try:
        res = walkfunc.compile2g(func_name, src_path, job_name=job_name)
        gbl.write_nl_lines(res, output_name)
        return 0

    except err.CompilerError as exn:

        for line in exn.get_report_lines():
            gbl.eprint(line)
        return 1


def prepare_optionns(options):
    # remove comments from source to docopt.
    parseable_opts = re.sub("\n# .*", "", DOC)
    # uncomment the maint options so they can
    # be parsed.
    parseable_opts = re.sub("\n!(.*)", "\n\\1", parseable_opts)
    opts = docopt.docopt(parseable_opts, help=False, argv=options)
    gbl.config = gbl.config._replace(
        bp_on_error=opts["--break"],
        short_filenames=opts["--short-filenames"],
        verbose=(int(opts["--verbose"])),
        narrow_output=gbl.config.narrow_output or opts["--narrow"],
        no_id=gbl.config.no_id or opts["--no-id"],
    )
    return opts


def do_help_options(opts):
    if opts["<topic>"]:
        do_doc(opts["<topic>"])
    elif opts["all"]:
        do_doc("all")
    elif opts["maint"]:
        docstr = re.sub("\n!", "\n", re.sub("\n[^!].*", "", DOC))
        gbl.sprint(docstr)
    elif opts["topics"]:
        do_doc("topics")
    else:
        # remove comment chars from usage
        # and the maint commands too.
        docstr = DOC.strip("\n").replace("#", "")
        docstr = re.sub("\n!.*", "", docstr)
        gbl.sprint(docstr)


def handled_dash_options(opts):
    if opts["version"]:
        from p2g import VERSION

        gbl.sprint(VERSION)
        return 1
    if opts["location"]:
        gbl.sprint(f"{__file__}")
        return 1
    if opts["help"]:
        do_help_options(opts)
        return 1
    if opts["examples"]:
        do_examples(pathlib.Path(opts["<dstfile>"]))
        return 1
    return 0


# trick - looking at the syntax for the options,
# if a 'srcfile' is called 'help', then we pretend were called
# with --help.  So you can't compile a file called help.
def trick_help_on_cmdline(opts):

    # if srcfile matches any of the other attributes, behave
    # as if that was typed.

    if opts["<srcfile>"] == "examples":
        opts["examples"] = True
        opts["<dstdir>"] = opts["<dstfile>"]
    elif opts["<srcfile>"] == "help":
        otherkey = opts["<dstfile>"]
        opts["help"] = True
        if otherkey in opts:
            opts[otherkey] = True
        else:
            opts["<topic>"] = otherkey


def main(options: typing.Optional[list[str]] = None):

    opts = prepare_optionns(options)

    trick_help_on_cmdline(opts)
    if handled_dash_options(opts):
        return 0

    return do_gen(
        src_name=opts["<srcfile>"],
        job_name=opts["--job"],
        func_name=opts["--function"],
        output_name=opts["<dstfile>"],
    )



import copy
import enum
import typing

from p2g import axis
from p2g import err
from p2g import gbl
from p2g import scalar
from p2g import stat
from p2g import vector


def flatten(args):
    res = []
    for el in args:
        for subel in el.everything():
            res.append(subel)
    return res


def unpack(args, kwargs) -> vector.TupleV:
    resmap: typing.Dict[int, scalar.Scalar] = {}
    if args:
        args = map(vector.wrap, args)
        args = flatten(args)
        for idx, value in enumerate(args):
            resmap[idx] = value
    # can have x=1,y=1 or xy=(somthing)
    for axis_string, values in kwargs.items():
        values = vector.wrap(values)
        axis_indexes = axis.name_to_indexes_list(axis_string)
        for axis_idx, value in zip(axis_indexes, values.forever()):
            if axis_idx in resmap:
                err.compiler(f"Overlapping axes {args} {kwargs}.")
            resmap[axis_idx] = value
    # fill any holes.
    coord_size = 0
    if resmap:
        coord_size = max(resmap) + 1
    # rules:
    # everything in the input gets packed into
    # the output in the right order.
    # if something is a nan in the input, and
    # we've got gap_default, then the nan is
    # turned into that.
    min_size = 0
    return vector.TupleV(
        [resmap.get(x) for x in range(max(min_size, coord_size))], from_user=True
    )


class CoNum(enum.IntEnum):
    CONST = enum.auto()
    FIXED = enum.auto()
    VAR = enum.auto()


# build things like AABuilder(1,2,3, _addr=123) and AABuilder[8](_addr=123)
class CoType:
    size: typing.Optional[int]
    cot: CoNum
    # for typing, default works too.
    x: "CoType"

    def __init__(self, cot: CoNum):
        self.cot = cot
        self.size = None

    def check_size_sanity(self, size, initialized_size):
        if initialized_size > 0 and initialized_size != size:
            raise err.CompilerError(
                f"Conflicting sizes {self.size} and {initialized_size}."
            )
        if size == 0 and initialized_size == 0:
            err.compiler("Zero sized vector.")

    def workout_size(self, initialized_size):
        size = self.size
        if size is None:
            size = max(initialized_size, 1)
        self.check_size_sanity(size, initialized_size)
        return size

    def __call__(self, *args, addr: typing.Optional[int] = None, **kwargs):
        values = unpack(args, kwargs)
        initialized_size = values.nelements()
        size = self.workout_size(initialized_size)
        match self.cot:
            case CoNum.CONST:
                if addr is not None:
                    raise err.CompilerError("Const can't have an address.")
                return values
            case CoNum.FIXED:
                if addr is None:
                    raise err.CompilerError("Fixed needs an address.")
            case CoNum.VAR:
                if addr is not None:
                    raise err.CompilerError("Var can't have an address.")
                addr = gbl.iface.next_bss(size)
            case _:  # no cover
                raise NotImplementedError
        res = vector.MemVec(addr, size)
        if initialized_size:
            for lhs, rhs in zip(res.everything(), values.forever()):
                stat.append_set(lhs, rhs)
        return res

    # sets vector size.
    def __getitem__(self, el):
        cop = copy.copy(self)
        cop.size = el
        return cop


Var = CoType(CoNum.VAR)
Fixed = CoType(CoNum.FIXED)
Const = CoType(CoNum.CONST)



#! /usr/bin/env python

import p2g
import usrlib as defs


def probecalibrate():

    p2g.comment(
        "Start with fixed height probe,",
        "make sure probe stickout <2.25in",
    )

    p2g.Control.symbol_table = True

    defs.goto_home()

    with defs.OTS():
        defs.pause("touch OTS, must beep")
        defs.goto_down(defs.MABS_ABOVE_OTS + defs.KNOWNLEN_OFF)
        defs.ots_calibrate()

    with defs.Probe():
        p2g.comment("Calibrate spindle probe.")
        defs.pause("touch probe, must beep")

        defs.goto_home()
        defs.goto_down(defs.MABS_ABOVE_OTS)
        defs.spindle_probe_find_height()

        p2g.comment("test spindle probe with ring.")
        defs.goto_home()
        defs.goto_down(defs.MABS_INSIDE_RING)
        defs.spindle_probe_find_radius(defs.PROBE_RING_DIAMETER)
        defs.goto_home()



import enum

import p2g


class Tool(enum.IntEnum):
    PROBE = 1
    KNOWN_LENGTH = 2


# pylint: disable=too-many-instance-attributes
# all coords are  abs end of spindle probe.
# ots tool and probe are different lengths.

KNOWNLEN_DIFF = -0.25
KNOWNLEN_OFF = p2g.Const(0, 0, KNOWNLEN_DIFF)
MABS_ABOVE_OTS = p2g.Const(x=-1.167, y=-7.56, z=-7.54)
MABS_INSIDE_RING = p2g.Const(x=-16.42, y=-3.42, z=-22.6)
MABS_ABOVE_ROTARY = p2g.Const(x=-12.5214, y=-12.9896, z=-7.0)
MABS_ABOVE_SEARCH_ROTARY_LHS_5X8 = p2g.Const(x=-15.5, y=-17.50, z=-14.0)
MABS_ABOVE_VICE = p2g.Const(x=-27.5, y=-13.0, z=-17.5)

PROBE_RING_DIAMETER = 0.7
FAST_FEED = 600.0
SLOW_PROBE_FEED = 10.0
FAST_PROBE_FEED = 60.0

must_skip_probe = p2g.goto.probe.mcode(p2g.haas.MUST_SKIP)
may_skip_probe = p2g.goto.probe

_slow_rel_probe = must_skip_probe.delay(0).relative.feed(SLOW_PROBE_FEED).all
_fast_rel_probe = must_skip_probe.relative.relative.all.feed(FAST_PROBE_FEED)
_fast_work_probe = may_skip_probe.work.feed(FAST_PROBE_FEED).all
_slow_work_probe = may_skip_probe.work.feed(SLOW_PROBE_FEED).all


def slow_rel_probe(**cos):
    _slow_rel_probe(**cos)


#    p2g.no_lookahead()


def slow_work_probe(**cos):
    _slow_work_probe(**cos)


#    p2g.no_lookahead()


def fast_rel_probe(**cos):
    _fast_rel_probe(**cos)


#    p2g.no_lookahead()


def fast_work_probe(**cos):
    _fast_work_probe(**cos)


#    p2g.no_lookahead()


goto_work = p2g.goto.work.feed(FAST_FEED).all
goto_G53 = p2g.goto.machine.feed(FAST_FEED)

# goto_down  - safeish way to go down,
# move xy first and then z.
goto_down = p2g.goto.work.feed(FAST_FEED).z_last

# goto_up  - safeish way to go up.
# move z first and then xy.

goto_up = p2g.goto.work.feed(FAST_FEED).z_first


goto_rel = p2g.goto.relative.feed(FAST_FEED).all
goto_work_z_first = p2g.goto.work.feed(FAST_FEED).z_first
goto = p2g.goto.work.feed(FAST_FEED).all


def goto_home():
    goto_G53.z_first(0, 0, 0)


def goto_zhome():
    goto_G53.all(z=0)


def setup_probing():
    load_tool(Tool.PROBE)
    p2g.codenl(p2g.haas.PROBE_ON, comment_txt=p2g.stat.CommentGen.FROMSRC)
    p2g.codenl(p2g.haas.NO_LOOKAHEAD, comment_txt=p2g.stat.CommentGen.NONE)


def pause(txt: str):
    p2g.comment()
    p2g.sys.message(p2g.haas.MESSAGE[0], txt)
    p2g.comment()


def alarm(txt: str):
    p2g.sys.message(p2g.haas.ALARM[0], txt)


def ots_calibrate():
    p2g.codenl("G65 P9023 A20. K5. S0.5 D-2.", comment_txt=p2g.stat.CommentGen.FROMSRC)


def spindle_probe_find_height():
    p2g.codenl("G65 P9023 A21. T1.", comment_txt=p2g.stat.CommentGen.FROMSRC)


def spindle_probe_find_radius(diameter):
    p2g.codenl(f"G65 P9023 A10.0 D{diameter}", comment_txt=p2g.stat.CommentGen.FROMSRC)


def load_tool(tool):
    p2g.codenl(f"T{tool:02} M06", comment_txt=p2g.stat.CommentGen.FROMSRC)


class Probe:
    def __enter__(self):
        load_tool(Tool.PROBE)
        p2g.codenl(p2g.haas.SPINDLE_PROBE_ON, comment_txt="Probe on.")

    def __exit__(self, *_):
        p2g.codenl(p2g.haas.SPINDLE_PROBE_OFF, comment_txt="Probe off.")


class OTS:
    def __enter__(self):
        load_tool(Tool.KNOWN_LENGTH)
        p2g.codenl(p2g.haas.OTS_ON, comment_txt="OTS On.")

    def __exit__(self, *_):
        p2g.codenl(p2g.haas.OTS_OFF, comment_txt="OTS Off.")


def optional_pause(txt: str):
    with p2g.sys.Optional():
        p2g.comment()
        pause(txt)
        p2g.comment()



from p2g import scalar


# Just type hints, can't be called.
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name


def abs(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


# def print(_v) -> scalar.ConstantNone:  # no cover
#     return scalar.ConstantNone()


def acos(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def asin(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def atan(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def cos(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def exp(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def fix(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def fup(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def ln(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def sin(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def sqrt(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def tan(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def exists(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()


def ground(_v) -> scalar.Scalar:  # no cover
    return scalar.ConstantNone()



import abc
import ast
import dataclasses
import typing
import string
from p2g import err
from p2g import gbl
from p2g import nd


WIDE_COMMENT_INDENT = 34
# infront of code which isn't optional block.
OPT_PREFIX = ""
# infront of code which isn't a label
NORMAL_PREFIX = "  "

# args to make comments


class CommentGen:  # pylint: disable=too-few-public-methods
    NONE = "no_comment"
    FROMSRC = "src_comment"
    FAIL = "FAIL"


# make translation maps from what we like
# to what the machine likes.  Like no parens
# in comments and so on
#
OK_CHARS_IN_COMMENTS = (
    string.ascii_lowercase
    + string.digits
    + "+-/* =.[]:,%_><\"'{}#&|^~!"
    + string.ascii_uppercase
)

OK_CHARS_IN_DPRNT = (
    string.ascii_uppercase
    + string.digits
    + "+-/*"
    + "[]"  # need these
    + string.ascii_lowercase  # these work for me, but aren't in the manual
    + ",#:.="
)


def setup_translations(okchars, **kwargs):
    translate_map = {
        **{chr(ch): "*" for ch in range(0, 127)},
        **{x: x for x in (okchars)},
        **kwargs,
    }

    return translate_map


@dataclasses.dataclass
class Label:
    idx: int
    used: bool

    def __init__(self, idx: int):
        self.idx = idx
        self.used = False

    def as_gcode_ref(self):
        return f"{self.idx}"

    def as_gcode_definition(self):
        return f"N{self.idx}"


class StatBase(abc.ABC):

    pos: ast.AST
    code_prefix: str

    translate_map = setup_translations(OK_CHARS_IN_COMMENTS, **{"(": "[", ")": "]"})

    @classmethod
    def clean_comment_chars(cls, txt: str):
        return "".join(cls.translate_map[ch] for ch in txt)

    # clean comment and remove some redundant
    # strings.
    @classmethod
    def compress_and_clean(cls, line: str):
        guts = cls.clean_comment_chars(line)
        for too_talky in ["p2g.", "var."]:
            guts = guts.replace(too_talky, "")
        if guts.startswith("    "):
            guts = guts[4:]
        return "( " + guts.ljust(30) + ")"

    def __init__(self, *, comtxt=CommentGen.FAIL):
        self.pos = gbl.iface.last_node
        self.comwant = comtxt
        self.code_prefix = gbl.Control.code_prefix

    def talkabout(self, orig):
        return f"{orig} {self.pos} {self.comwant} {self.code_prefix}"

    @gbl.g2l
    def to_line_lhs(self):
        yield "never"  # no cover

    @gbl.g2l
    # asks derived classes to provide code lhs,
    # and then fills in with the common case comment.
    def to_full_lines(self, nest):
        match self.comwant:
            case CommentGen.FROMSRC:
                comtxt = self.compress_and_clean(err.source_from_node(self.pos))
            case CommentGen.NONE:
                comtxt = CommentGen.NONE
            case txt:
                comtxt = self.compress_and_clean(txt)

        # only print a comment once.
        if comtxt == nest.prev_comtxt:
            comtxt = CommentGen.NONE
        else:
            nest.prev_comtxt = comtxt

        code_gen = self.to_line_lhs()

        # code_list is a sequence of gcode, com_txt is a comment
        # which is applied to first line if it can, otherwise
        # is yielded on own line.
        try:
            chunk = next(code_gen)
            prefix_and_code = self.code_prefix + chunk
            code_comment_gap = WIDE_COMMENT_INDENT - len(prefix_and_code)
            indented_comment = " " * code_comment_gap + comtxt

            if comtxt == CommentGen.NONE:
                yield prefix_and_code
            elif gbl.config.narrow_output or code_comment_gap < 0:
                yield comtxt
                yield prefix_and_code
            else:
                yield prefix_and_code + indented_comment

            yield from (self.code_prefix + chunk for chunk in code_gen)
        except StopIteration as exn:  # no cover
            raise AssertionError from exn


class EmacsClientState:
    last_file: str
    last_line: str

    def __init__(self):
        self.last_file = ""
        self.last_line = ""


@dataclasses.dataclass
class Nest:
    first_label: int
    next_label: int
    prev_comtxt: str
    slist: list[StatBase]
    cur: typing.Optional["Nest"] = None
    ecs: EmacsClientState = EmacsClientState()

    def __init__(self):
        self.prev_comtxt = ""
        self.first_label = 1000
        self.next_label = self.first_label
        self.slist = []

    @classmethod
    def get_label(cls):
        assert cls.cur is not None
        res = cls.cur.next_label
        cls.cur.next_label += 1
        return Label(res)

    @classmethod
    def add_stat(cls, stat):

        gbl.v2print(f"Adding {stat}")
        if Nest.cur is None:
            return
        Nest.cur.slist.append(stat)

    def to_full_lines(self):
        for line in self.slist:
            yield from line.to_full_lines(self)

    @classmethod
    def __enter__(cls):
        assert cls.cur is None
        cls.cur = Nest()
        return cls.cur

    @classmethod
    def __exit__(cls, *_exn):
        cls.cur = None


class CommentLines(StatBase):
    def __init__(self, lines):
        super().__init__(comtxt=CommentGen.NONE)
        self.lines = lines

    def to_full_lines(self, _):
        lines = list(map(str, self.lines))

        def max_str_len(lines):
            return len(max(lines, key=len, default=""))

        def pad_to_same_width(lines):
            wid = max_str_len(lines)
            res = []
            for line in lines:
                res.append(line.ljust(wid))
            return res

        lines = pad_to_same_width(lines)
        for line in lines:
            if line.strip():
                yield "( " + StatBase.clean_comment_chars(line) + " )"
            else:
                yield ""


add_stat = Nest.add_stat


def comment(*lines):
    add_stat(CommentLines(" "))
    add_stat(CommentLines(lines))


def com(*lines):
    add_stat(CommentLines(lines))


@dataclasses.dataclass
class Goto(StatBase):
    target: Label

    def __init__(self, target: Label):
        super().__init__(comtxt=CommentGen.NONE)
        self.target = target

    def to_line_lhs(self):
        yield f"{NORMAL_PREFIX}GOTO {self.target.as_gcode_ref()}"


@dataclasses.dataclass
class If(StatBase):
    exp: typing.Any
    on_t: Label

    def __init__(self, exp, on_t: Label):
        super().__init__(comtxt=CommentGen.FROMSRC)
        self.exp = exp
        self.on_t = on_t

    def to_line_lhs(self):
        lhs = f"IF [{nd.to_gcode(self.exp)}] "
        rhs = f"GOTO {self.on_t.as_gcode_ref()}"
        yield f"{NORMAL_PREFIX}{lhs}{rhs}"


@dataclasses.dataclass
class IfSet(StatBase):
    cond: typing.Any
    ass: "Set"

    def __init__(self, cond, ass, comment_txt=CommentGen.FROMSRC):
        super().__init__(comtxt=comment_txt)
        self.cond = cond
        self.ass = ass

    def to_line_lhs(self):
        yield f"{NORMAL_PREFIX}IF [{nd.to_gcode(self.cond)}] {self.ass}"


class Code(StatBase):
    txt: str

    def __init__(self, txtargs, comment_txt=CommentGen.FROMSRC):
        super().__init__(comtxt=comment_txt)
        self.txt = gbl.unwind(txtargs)

    def to_line_lhs(self):
        if self.txt and (self.txt[0] in "O%LN"):
            yield self.txt
        else:
            yield f"{NORMAL_PREFIX}{self.txt}"

    def __repr__(self):
        return self.talkabout(f"Code: {self.txt}")


@dataclasses.dataclass
class Dprint(StatBase):
    txt: str
    translate_map = setup_translations(
        OK_CHARS_IN_DPRNT,
        #        **{x: x.upper() for x in (string.ascii_lowercase)},
    )

    def __init__(self, txt):
        super().__init__(comtxt=CommentGen.NONE)

        self.txt = "".join(Dprint.translate_map[ch] for ch in txt)

    def to_line_lhs(self):

        yield "DPRNT[" + self.txt + "]"


class Set(StatBase):
    args: list[nd.EBase]
    msg_txt: str

    def __init__(
        self,
        lhs: nd.EBase,
        rhs: nd.EBase,
        *,
        msg_txt: str = "",
        comment_txt=CommentGen.FROMSRC,
    ):
        super().__init__(comtxt=comment_txt)
        assert isinstance(lhs, nd.EBase)
        self.args = [lhs, rhs]
        self.msg_txt = msg_txt

    def to_line_lhs(self):
        yield f"{NORMAL_PREFIX}{self}"

    def __repr__(self):
        lhs = nd.to_gcode(self.args[0])
        rhs = nd.to_gcode(self.args[1], modifier=nd.NodeModifier.EMPTY)
        msg_txt = (
            " ( " + self.clean_comment_chars(self.msg_txt) + " )" if self.msg_txt else ""
        )
        return f"{lhs}= {rhs}{msg_txt}"


def append_set(dst, src, comment_txt=CommentGen.FROMSRC):
    if not gbl.same(dst, src):
        add_stat(Set(dst, src, comment_txt=comment_txt))


def code(*txtargs, comment_txt=CommentGen.FROMSRC):
    add_stat(Code(txtargs, comment_txt))


def codenl(txtlst, comment_txt=CommentGen.FROMSRC):
    if isinstance(txtlst, str):
        add_stat(Code(txtlst, comment_txt))
    else:
        for txt in txtlst:
            codenl(txt, comment_txt)
            comment_txt = CommentGen.NONE


# contents evaluate when needed, not when
# created.
class Lazy(StatBase):
    todo: typing.Generator[str, None, None]

    def __init__(self, todo: typing.Generator[str, None, None]):
        super().__init__(comtxt=CommentGen.NONE)
        self.todo = todo

    def to_full_lines(self, _):
        if self.todo:
            yield from self.todo


@dataclasses.dataclass
class LabelDef(StatBase):
    labeldef: Label

    def __init__(self, labeldef: Label):
        self.labeldef = labeldef
        super().__init__(comtxt=CommentGen.NONE)

    def to_line_lhs(self):
        yield f"{self.labeldef.as_gcode_definition()}"


@dataclasses.dataclass
class Percent(StatBase):
    def __init__(self):
        super().__init__(comtxt=CommentGen.NONE)

    def to_line_lhs(self):
        yield "%"


next_label = Nest.get_label



import ast
import contextlib
import functools
import pathlib
import sys
import typing


class Config(typing.NamedTuple):
    narrow_output: bool = False
    bp_on_error: bool = False
    verbose: int = 0
    no_id: bool = False
    short_filenames: bool = False
    in_pytestwant: bool = False
    first_line: int = -1


config = Config()

# pylint: disable=too-few-public-methods


class Control:
    code_prefix = ""
    symbol_table = False
    emacsclient = False


def log(*args):
    if config.verbose > 1:
        print(*args)


def v1print(*args):
    if config.verbose > 0:
        print(*args)


def v2print(*args):
    if config.verbose > 1:
        print(*args)


def sprint(*args):
    print(*args)


# print to redirected stderr
def eprint(*args):
    print(*args, file=sys.stderr)


######################################################################


# needs to be cacheable cause stdin is possible.
# needs to be clearable because get called with stdin
# with different contents


@functools.cache
def get_lines(name):
    # pylint: disable=consider-using-with
    if str(name) == "-":
        file = sys.stdin
    else:
        file = open(name, encoding="utf-8")

    return file.read().split("\n")

    # print("IN GETLINES ", name)
    # # parse command line
    # if str(name) != "-":
    #     inf = open(name, encoding="utf-8")
    #     guts = inf.read().split("\n")

    # else:
    #     guts = sys.stdin.read().split("\n")
    #     print("READ", guts)

    # return guts


def g2l(generator):
    def g2l_(*args, **kwargs):
        return list(generator(*args, **kwargs))

    return g2l_


# generator returning one string from a yield of many
def g2s(generator):
    def g2s_(*args, **kwargs):
        return "".join(generator(*args, **kwargs))

    return g2s_


def sentinel():
    # pylint: disable=too-few-public-methods
    class Sentinel:
        pass

    return Sentinel


# open accouting for '-' files.
# sourcing empty file when in test.
@contextlib.contextmanager
def openw(name):
    if str(name) == "-":
        yield sys.stdout
    else:
        with open(name, "w", encoding="utf-8") as wfile:
            yield wfile


def write_nl_lines(thing, outfile_path):
    with openw(outfile_path) as outf:
        for line in thing:
            print(line, file=outf)


def same(lhs, rhs):
    if type(lhs) is not type(rhs):
        return False
    return lhs.same(rhs)


# unwind a list of lists etc can turn generators of lists of strings
# into space joined string.
def unwind1(args):
    if isinstance(args, str):
        yield args
    else:
        for el in args:
            yield from unwind1(el)


def unwind(args):
    return " ".join(unwind1(args))


# find the file if in dist or just checkout out.
def find_ours(filename):
    for topdir in [".", ".."]:
        for subdir in ["doc", "examples"]:
            look_here = pathlib.Path(__file__).parent / topdir / subdir / filename
            if look_here.exists():
                return look_here.resolve()
    raise FileNotFoundError  # no cover


on_exit: typing.List[typing.Callable] = []


def set_ast_file_name(node: ast.AST, name):

    setattr(node, "file_name", name)


def ast_file_name(node: ast.AST):
    return getattr(node, "file_name")


def make_fake_node(filename, lineno, offset, end_offset):
    fake_node = ast.AST()
    set_ast_file_name(fake_node, filename)
    fake_node.lineno = lineno
    fake_node.col_offset = offset
    fake_node.end_col_offset = end_offset
    return fake_node


astnone = make_fake_node("empty", 1, 0, 0)


class PerTranslation:
    ebss: int
    varrefs: typing.Dict
    last_node: ast.AST

    def __init__(self):
        self.ebss = 100

        self.varrefs = {}
        self.last_node = astnone

    def next_bss(self, size):
        addr = self.ebss
        self.ebss += size
        return addr


iface: "PerTranslation" = PerTranslation()


def reset():
    #
    # pylint: disable=global-statement
    global iface
    get_lines.cache_clear()
    iface = PerTranslation()



from p2g import err
from p2g import scalar


NAMES = "xyz"
# special case constant for vectors of naxes size.


class Axes(scalar.Constant):
    def __init__(self):
        super().__init__(len(NAMES))

    @property
    def _value(self):
        return len(NAMES)

    @_value.setter
    def _value(self, _inset):
        pass


def low_names_v():
    return NAMES.lower()


def low_names(idx):
    return NAMES[idx]


def hi_names(idx):
    return NAMES[idx].upper()


def axis_name_to_index(axis_char):
    return NAMES.index(axis_char)


def name_to_indexes_list(name: str) -> list[int]:
    try:
        return [axis_name_to_index(ch) for ch in name]
    except ValueError:
        return err.compiler(f"Bad axis letter in '{name}'.")



import ast
import pathlib

from p2g import gbl


def code_context_node(node: ast.AST):

    file_name = gbl.ast_file_name(node)
    file_path = pathlib.Path(file_name).absolute()
    reportable_line = node.lineno

    # if testing try and shrink filename and make
    # line numbers relative to start of function.

    if gbl.config.first_line >= 0:
        reportable_line -= gbl.config.first_line
    if gbl.config.short_filenames:
        file_name = str(file_path.name)
    else:
        file_name = str(file_path)

    col_offset = node.col_offset
    assert isinstance(node.end_col_offset, int)
    assert isinstance(node.col_offset, int)
    assert node.end_col_offset is not None

    col_width = node.end_col_offset - col_offset
    pfx = ":".join(
        [
            file_name,
            str(reportable_line),
            str(node.col_offset),
            str(node.end_col_offset),
            " ",
        ]
    )
    return pfx, " " * (len(pfx) + col_offset) + "^" * col_width


def source_from_node(node: ast.AST):
    source_lines = gbl.get_lines(gbl.ast_file_name(node))
    return source_lines[node.lineno - 1]


class CompilerError(Exception):

    node: ast.AST
    message: str
    report_line: bool

    def __init__(self, message, report_line=True, node=gbl.astnone):
        super().__init__()

        if gbl.config.bp_on_error:
            breakpoint()  # no cover

        self.report_line = report_line
        self.message = message
        self.node = gbl.iface.last_node if node is gbl.astnone else node

    def get_report_lines(self):

        node = self.node
        message = self.message

        line_prefix, source_context = code_context_node(node)

        if not self.report_line:
            return [source_context, message]

        source_line = source_from_node(node)
        # if error messaeg text will fit into context
        # line without trashing anything then put it there.
        message_len = len(message)
        gap_to_start = len(source_context) - len(source_context.lstrip())

        space_between = gap_to_start - message_len
        if space_between > 1:
            source_context = message + " " * space_between + source_context[gap_to_start:]
            return [line_prefix + source_line, source_context]

        return [line_prefix + source_line, source_context, message]


def compiler(message: str = "", node=gbl.astnone, exn=None):

    raise CompilerError(message, node=node) from exn



import ast
import dataclasses

from p2g import coords
from p2g import err
from p2g import nd
from p2g import op
from p2g import scalar
from p2g import stat
from p2g import vector
from p2g import walkbase


# wrap enough state that we can call an
# ast function as if it were native.
def ensure_no_more(src_gen):
    try:
        next(src_gen)
        err.compiler("Too many values to unpack.")
    except StopIteration:
        pass


def handle_assign(self, target, val):
    if not isinstance(target, ast.Tuple):
        self.visit_store(target, val)
        return
    src_gen = iter(val)  # pytype: disable=wrong-arg-types
    try:
        for dst_idx, dst_el in enumerate(target.elts):
            if isinstance(dst_el, ast.Starred):
                src_togo = list(src_gen)
                break_i = len(src_togo) - (len(target.elts) - dst_idx - 1)
                self.visit_store(dst_el.value, src_togo[:break_i])
                src_gen = iter(src_togo[break_i:])
            else:
                self.visit_store(dst_el, next(src_gen))
    except StopIteration:
        err.compiler("Not enough values to unpack.")
    ensure_no_more(src_gen)


# make code for while and for loops.
@dataclasses.dataclass
class ItrRange:
    target: ast.AST

    def __init__(self, itr, interp, target):
        self.interp = interp
        self.rang = itr
        self.var = coords.Var()
        if not isinstance(target, ast.Name):
            err.compiler("Must be simple name as destination for 'for'.")
        self.var = coords.Var()
        interp.ns[target.id] = self.var

    #        interp.ns[target.id] = self.var
    def setup_index(self):
        pass

    def setup_target(self):
        stat.add_stat(
            stat.Set(scalar.wrap_scalar(self.var), scalar.wrap_scalar(self.rang.start))
        )

    def cur_index(self):
        return self.var

    def past_index(self):
        return self.rang.stop

    def load_target(self):
        pass

    def next_index(self):
        stat.add_stat(
            stat.Set(
                scalar.wrap_scalar(self.var),
                op.make_scalar_add(scalar.wrap_scalar(self.var), self.rang.step),
                comment_txt=stat.CommentGen.NONE,
            )
        )


@dataclasses.dataclass
class ItrSlice:
    ptr: nd.EBase
    pastptr: scalar.Scalar
    step: scalar.Scalar
    var: ast.AST

    def __init__(self, itr, interp, target):
        self.interp = interp
        self.target = target
        # with a slice, always need a pointer to
        # look at things.
        # scalar.wrap_scalar(coords.Var(itr._addr))
        self.ptr = coords.Var(itr._addr)
        if not isinstance(target, ast.Name):
            err.compiler("Must be simple name as destination for 'for'.")
        self.pastptr = itr._addr + itr._size
        self.step = itr._step

    def setup_index(self):
        pass

    def setup_target(self):
        pass

    def cur_index(self):
        return self.ptr

    def past_index(self):
        return self.pastptr

    def load_target(self):
        handle_assign(self.interp, self.target, op.hashop(scalar.wrap_scalar(self.ptr)))

    def next_index(self):
        stat.add_stat(
            stat.Set(
                scalar.wrap_scalar(self.ptr),
                op.make_scalar_add(scalar.wrap_scalar(self.ptr), self.step),
                comment_txt=stat.CommentGen.NONE,
            )
        )


def make_itr(itr, interp, target):
    if isinstance(itr, vector.Vec):
        return ItrSlice(itr, interp, target)
    if isinstance(itr, range):
        return ItrRange(itr, interp, target)
    raise err.CompilerError("Illegal iterator.")


def geniter(
    self,
    *,
    whileexp=None,
    target=None,
    itr=None,
    body=None,
    orelse=None,
):
    self.loop = walkbase.LoopContext(self.loop)
    if whileexp:
        stat.add_stat(stat.LabelDef(self.loop.lcontinue))
        test = op.make_scalar_unop(op.a2opfo(ast.Not), self.visit(whileexp))
        stat.add_stat(stat.If(test, self.loop.lorelse))
        self.visit_slist(body)
        stat.add_stat(
            stat.Goto(
                self.loop.lcontinue,
            ),
        )
    else:
        assert itr
        itr = make_itr(itr, self, target)
        itr.setup_index()
        itr.setup_target()
        stat.add_stat(stat.LabelDef(self.loop.lcontinue))
        stat.add_stat(
            stat.If(
                op.make_vec_binop(
                    op.a2opfo(ast.GtE),
                    itr.cur_index(),
                    itr.past_index(),
                ),
                self.loop.lorelse,
            )
        )
        itr.load_target()
        self.visit_slist(body)

        itr.next_index()
        stat.add_stat(
            stat.Goto(
                self.loop.lcontinue,
            ),
        )
    if orelse:
        stat.add_stat(stat.LabelDef(self.loop.lorelse))
        self.visit_slist(orelse)
        if self.loop.lbreak.used:
            stat.add_stat(stat.LabelDef(self.loop.lbreak))
    else:
        if self.loop.lbreak.used:
            stat.add_stat(stat.LabelDef(self.loop.lbreak))
        stat.add_stat(stat.LabelDef(self.loop.lorelse))
    self.loop = self.loop.prev


class WalkStatement(walkbase.WalkBase):
    def wrap_decorators(self, obj, node):
        for deco_n in reversed(node.decorator_list):
            deco = self.visit(deco_n)
            obj = deco(obj)
        return obj

    def _visit_module(self, node, file_name):
        self.ns = self.module_ns = walkbase.Namespace()
        self.ns["__file__"] = file_name
        self.file_name = file_name
        self.ns["__name__"] = "__main__"
        self.visit_slist(node.body)

    def visit_module(self, node, file_name):
        self._visit_module(node, file_name)

    def _visit_classdef(self, node):
        with self.pushpop_classns() as ns:
            self.visit_slist(node.body)
            guts = ns.guts
        cls = type(node.name, tuple(self.visit(b) for b in node.bases), guts)
        cls = self.wrap_decorators(cls, node)
        self.ns[node.name] = cls

    def _visit_for(self, node):
        itr = self.visit(node.iter)
        return geniter(
            self, target=node.target, itr=itr, body=node.body, orelse=node.orelse
        )

    def _visit_while(self, node):
        geniter(self, whileexp=node.test, body=node.body, orelse=node.orelse)

    def _visit_break(self, _):
        self.loop.lbreak.used = True
        stat.add_stat(stat.Goto(self.loop.lbreak))

    def _visit_continue(self, _):
        stat.add_stat(stat.Goto(self.loop.lcontinue))

    def _visit_with(self, node):
        ctxes = [self.visit(el.context_expr) for el in node.items]
        entry_vals = (ctx.__enter__() for ctx in ctxes)
        for el, val in zip(node.items, entry_vals):
            if var := el.optional_vars:
                handle_assign(self, var, val)
        self.visit_slist(node.body)
        for ctx in reversed(ctxes):
            ctx.__exit__(None, None, None)

    def _visit_if(self, node):
        #   for nicer looking code
        # for a conditional break.
        if isinstance(node.body[0], ast.Break):
            exp = op.make_scalar_unop(op.a2opfo(ast.UAdd), self.visit(node.test))
            self.loop.lbreak.used = True
            stat.add_stat(stat.If(exp, self.loop.lbreak))
            return
        elsepart = stat.next_label()
        donepart = stat.next_label()
        cond = op.make_scalar_unop(op.a2opfo(ast.UAdd), self.visit(node.test))
        # nice code for if with constant.
        if cond.is_constant:
            if cond:
                self.visit_slist(node.body)
            else:
                self.visit_slist(node.orelse)
        else:
            exp1 = op.make_scalar_unop(op.a2opfo(ast.Not), cond)
            stat.add_stat(stat.If(exp1, on_t=elsepart))
            self.visit_slist(node.body)
            stat.add_stat(
                stat.Goto(
                    donepart,
                ),
            )
            stat.add_stat(
                stat.LabelDef(
                    elsepart,
                ),
            )
            self.visit_slist(node.orelse)
            stat.add_stat(
                stat.LabelDef(
                    donepart,
                ),
            )

    def _visit_import(self, node):
        for name in node.names:
            self.ns[name.asname or name.name] = __import__(name.name)

    def _visit_importfrom(self, node):
        mod = __import__(
            node.module, None, None, [n.name for n in node.names], node.level
        )
        for name in node.names:
            if name.name == "*":
                for modname, modguts in mod.__dict__.items():
                    if modname[0] != "_":
                        self.ns[modname] = modguts
            else:
                self.ns[name.asname or name.name] = getattr(mod, name.name)

    def _visit_augassign(self, node):
        self.visit_store(
            node.target,
            op.make_vec_binop(
                op.a2opfo(type(node.op)),
                self.visit(node.target),
                self.visit(node.value),
            ),
        )

    def _visit_annassign(self, node):
        if node.value is not None:
            val = self.visit(node.value)
            handle_assign(self, node.target, val)

    def _visit_assign(self, node):
        val = self.visit(node.value)
        for dst in node.targets:
            handle_assign(self, dst, val)

    def _visit_delete(self, node):
        for dst in node.targets:
            self.visit(dst)

    def _visit_pass(self, _):
        pass

    # assert turns into if with alarm.
    def _visit_assert(self, node):
        assert_ok = scalar.wrap_scalar(self.visit(node.test))
        msg_txt = self.visit(node.msg)
        ass_stat = stat.Set(
            scalar.wrap_scalar(3001),
            scalar.Constant(100),
            msg_txt=msg_txt,
        )
        if assert_ok.is_constant:
            if assert_ok:
                stat.add_stat(stat.Code(["( ok )"]))
                return
            stat.add_stat(ass_stat)
            return
        ifexp = op.make_scalar_unop(op.a2opfo(ast.Not), assert_ok)
        stat.add_stat(stat.IfSet(ifexp, ass_stat))



import ast
import dataclasses
import itertools
import pathlib
import re
import sys as realsys
import typing

import p2g

from p2g import axis
from p2g import err
from p2g import gbl
from p2g import op
from p2g import stat
from p2g import symbol
from p2g import sys
from p2g import walkbase
from p2g import walkexpr
from p2g import walkstat


# unique type to compare
Marker = gbl.sentinel()
# look through caller's args and insert defaults
# etc. to be compatible with callee, return dict
# with mapping.


def formal_kwargs(formalspec, formals, kwargs, pos) -> dict[str, typing.Any]:

    # Process incoming keyword arguments, putting them in namespace if
    # actual arg exists by that name, or offload to function's kwarg
    # if any. All make needed checks and error out.
    func_kwarg = {}
    final_dict = {}
    all_formals = {
        x.arg
        for x in itertools.chain(formals.args, formals.kwonlyargs, formals.kw_defaults)
        if x is not None
    }
    for key, value in kwargs.items():
        if key in all_formals:
            final_dict[key] = value
        elif formalspec.kwarg:
            func_kwarg[key] = value
        else:
            err.compiler("Bad arguments.", node=pos)
    if formalspec.kwarg:
        final_dict[formalspec.kwarg.arg] = func_kwarg
    return final_dict


def check_missing(final_dict, formalspec, pos):
    # check for missing args
    for formal in itertools.chain(formalspec.args, formalspec.kwonlyargs):
        if formal.arg not in final_dict:
            err.compiler(f"Missing argument '{formal.arg}'.", node=pos)


def get_defaults(walker, formals):
    first_defaulted = len(formals.args) - len(formals.defaults)
    return {
        # position args defaults
        el.arg: walker.visit(formals.defaults[idx - first_defaulted])
        for idx, el in enumerate(formals.args)
        if idx >= first_defaulted
    } | {
        # kwargs defaults
        el.arg: walker.visit(val)
        for el, val in zip(formals.kwonlyargs, formals.kw_defaults)
        if val is not None
    }


def gather_func_formals(func_def, *args, **kwargs):
    walker = func_def.walker
    formals = func_def.node.args
    # report error in caller rather than definition.

    pos = gbl.iface.last_node
    final_dict = {}
    formalspec = func_def.node.args
    if formalspec.vararg:
        final_dict[formalspec.vararg.arg] = args[len(formalspec.args) :]
    else:
        if len(args) > len(formalspec.args):
            err.compiler(
                f"Too many arguments; {len (args)} > {len(formalspec.args)}.", node=pos
            )
    for argidx in range(min(len(args), len(formalspec.args))):
        final_dict[formalspec.args[argidx].arg] = args[argidx]
    # result contains at least the defaults.
    final_dict |= get_defaults(walker, formals)
    final_dict |= formal_kwargs(formalspec, formals, kwargs, pos)
    check_missing(final_dict, formalspec, pos)
    return final_dict


# calling a function, emit code to call, and
# generate the called code.  And save it, to make
# sure that other generations make the same.
def inline(func_def, *args, **kwargs):

    walker = func_def.walker

    prev_file = walker.file_name
    prev_func = walker.func_name
    walker.file_name = func_def.file_name
    walker.func_name = func_def.func_name

    # We need to switch from dynamic execution scope to lexical scope
    # in which function was defined (then switch back on return).
    dyna_scope = walker.ns

    with walker.pushpopns(func_def.lexical_scope):
        assert walker.ns == func_def.lexical_scope
        res = None
        with walker.pushpop_funcns():
            formals_dict = gather_func_formals(func_def, *args, **kwargs)
            walker.ns.guts.update(formals_dict)
            res = walker.visit_slist(func_def.node.body)

    assert walker.ns == dyna_scope
    walker.func_name = prev_func
    walker.file_name = prev_file
    return res


@dataclasses.dataclass
class FuncDefWrap:
    file_name: str
    func_name: str
    node: ast.AST
    walker: "WalkFunc"
    lexical_scope: walkbase.Namespace

    def __init__(self, walker, node):
        self.func_name = node.name
        self.file_name = walker.module_ns["__file__"]
        self.node = node
        self.walker = walker
        self.lexical_scope = walker.ns

    def __call__(self, *args, **kwargs):
        return inline(self, *args, **kwargs)


def interpfunc(fun):
    def func(*args, **kwargs):
        if args and isinstance(args[0], Marker):
            return fun

        return fun(*args, **kwargs)

    return func


@dataclasses.dataclass
class FuncArgsDescr:
    args: list[typing.Any]
    kwargs: dict[str, typing.Any]

    def __init__(self, walker, node):
        args = []
        for arg in node.args:
            if isinstance(arg, ast.Starred):
                args.extend(walker.visit(arg.value))
            else:
                args.append(walker.visit(arg))
        kwargs = {}
        for keyword in node.keywords:
            val = walker.visit(keyword.value)
            if keyword.arg is None:
                kwargs.update(val)
            else:
                kwargs[keyword.arg] = val
        self.args = args
        self.kwargs = kwargs


def funcall(target, *args, **kwargs):
    return target(*args, **kwargs)


class WalkFunc(walkbase.WalkBase):
    def _visit_call(self, node):
        defn = self.visit(node.func)
        if defn.__module__ == "p2g.sys" and defn.__name__ == "print":
            with sys.BSS():
                # keep bss so 'stack' used gathering dprint arguments
                # gets returned.
                desc = FuncArgsDescr(self, node)
                return funcall(defn, *desc.args, **desc.kwargs)

        desc = FuncArgsDescr(self, node)

        if defn.__module__ == "p2g.builtin":
            return op.make_scalar_func(defn.__name__, *desc.args)
        return funcall(defn, *desc.args, **desc.kwargs)

    def _visit_functiondef(self, node):
        desc = FuncDefWrap(self, node)
        ifunc = interpfunc(desc)
        self.ns[node.name] = ifunc


def find_desc(walker, func_name, srcpath):
    try:
        fncdef = walker.ns[func_name]
        desc = fncdef(Marker())
    except KeyError as exn:
        raise err.CompilerError(
            f"No such function '{func_name}' in '{srcpath}'."
        ) from exn
    return desc


class Walk(
    walkstat.WalkStatement,
    walkexpr.WalkExpr,
    walkbase.WalkNS,
    WalkFunc,
    walkbase.WalkBase,
):
    pass


def find_defined_funcs(sourcelines):
    for line in sourcelines:
        # find last line with def in it, that's the function we need
        mares = re.match("def (.*?)\\(", line)
        if mares:
            yield mares.group(1)


# make sure any def test_func is after the TESTS BELOW
# comment, or future sedding will make us sad.
def check_test_after_marker(sourcelines, node):
    if not gbl.config.in_pytestwant:
        return

    had_marker = False
    for line in sourcelines:
        if line.startswith("# TESTS BELOW"):
            had_marker = True
        if line.startswith("def test_"):
            if not had_marker:
                err.compiler(f"need TEST BELOW before {line}", node=node)


def find_main_func_name(sourcelines, func_name_arg):
    if func_name_arg != "<last>":
        return func_name_arg

    function_to_call = "no function in file"

    for fname in find_defined_funcs(sourcelines):
        function_to_call = fname

    return function_to_call


def digest_top(walker, func_name, srcpath):
    desc = find_desc(walker, func_name, srcpath)
    stat.add_stat(stat.Lazy(symbol.Table.yield_table()))

    inline(desc)
    stat.codenl(["M30"], comment_txt=stat.CommentGen.NONE)
    for handler in gbl.on_exit:
        handler()

    stat.add_stat(stat.Percent())
    #    stat.codenl(["%"], comment_txt=stat.CommentGen.NONE)
    return desc


@gbl.g2l
def compile2g(func_name_arg, srcfile_name, job_name):

    with stat.Nest() as cursor:
        try:
            axis.NAMES = "xyz"
            gbl.reset()
            symbol.Table.reset()

            src_path = pathlib.Path(srcfile_name)
            src_lines = gbl.get_lines(src_path)

            realsys.path.insert(0, str(src_path.parent))

            gbl.log(f"Starting {func_name_arg} {cursor.next_label}")
            func_name = find_main_func_name(src_lines, func_name_arg)

            version = "" if gbl.config.no_id else f": {p2g.VERSION}"
            stat.code(
                f"{job_name} ({func_name}{version})",
                comment_txt=stat.CommentGen.NONE,
            )

            node = ast.parse("\n".join(src_lines), filename=srcfile_name)

            for el in ast.walk(node):
                gbl.set_ast_file_name(el, srcfile_name)

            # load everything

            walker = Walk()
            walker.visit_module(node, srcfile_name)

            if node.body:
                funcdef = digest_top(
                    walker,
                    func_name,
                    src_path,
                )

                check_test_after_marker(src_lines, funcdef.node)

            # careful with use of generators, because symbol table may
            # be emitted at the top, yet needs things used last on.
            res = list(cursor.to_full_lines())

            yield from res
        except FileNotFoundError as exn:
            # happens when python can't open file
            togo_node = gbl.make_fake_node(srcfile_name, 0, 0, 0)
            raise err.CompilerError(str(exn), report_line=False, node=togo_node) from exn
        except SyntaxError as exn:
            # comes from inside python when importing file.

            togo_node = gbl.make_fake_node(
                exn.filename, exn.lineno, exn.offset, exn.end_offset
            )
            raise err.CompilerError(exn.msg, node=togo_node) from exn

        except (
            TypeError,
            KeyError,
            ModuleNotFoundError,
            AttributeError,
            IndexError,
        ) as exn:
            raise err.CompilerError(str(exn)) from exn



# just information about operators.
import abc
import ast
import enum
import typing


def opt_null(*_):
    return None


class Opinfo(typing.NamedTuple):
    astc: typing.Optional[typing.Type] = None
    pyn: str = ""
    mth: str = ""
    lam: typing.Callable = opt_null
    gname: str = ""
    prec: int = 20
    comm: bool = False
    nargs: int = 2
    g_func: bool = False
    opt: typing.Callable = opt_null

    # def rtl_get_arg_(self, _idx):
    #     return self.pyn

    # def rtl_arg_info_(self):
    #     return ["opinfo"]


const_nd = Opinfo(astc=ast.Constant, pyn="konstant", gname="", prec=20)


class NodeModifier(enum.IntFlag):
    # so 0 is an error
    EMPTY = enum.auto()
    # ADDRESSES are always integers
    ADDRESS = enum.auto()
    ARGUMENT = enum.auto()
    NOSPACE = enum.auto()
    # TAKES UP 7 spaces.
    F3X3 = enum.auto()


# operators are generated mechanically, but these
# are for typing hints:
class FakeOps:
    def __div__(self, _other):
        return self  # no cover

    def __add__(self, _other):
        return self  # no cover

    def __sub__(self, _other):
        return self  # no cover

    def __mul__(self, _other):
        return self  # no cover

    def __floordiv__(self, _other):
        return self  # no cover

    def __neg__(self):
        return self  # no cover

    def __abs__(self):
        return self  # no cover

    def __lt__(self, _other):
        return self  # no cover

    def __gt__(self, _other):
        return self  # no cover


# common base class for scalar and vector.
class EBase(abc.ABC, FakeOps):
    @abc.abstractmethod
    def everything(self) -> typing.Generator[typing.Any, None, None]:
        ...

    def get_address(self):
        raise NotImplementedError

    def set_at(self, _idx, _src):
        raise NotImplementedError

    # just for typing sanity.
    def __div__(self, other):
        pass  # no cover

    def __mult__(self, other):
        pass  # no cover


DECIMALS = 4


def to_gcode_from_float(thing, modifier=NodeModifier.EMPTY):
    if modifier & NodeModifier.ADDRESS:
        return str(int(thing))
    if modifier & NodeModifier.F3X3:
        return f"{thing:7.3f}"
    res = str(round(float(thing), DECIMALS))
    if res.endswith(".0"):
        res = res[:-1]
    return res


def to_gcode(thing, modifier=NodeModifier.EMPTY) -> str:
    if isinstance(thing, (float, int)):
        return to_gcode_from_float(thing, modifier)

    return thing.to_gcode(modifier)



import ast

from p2g import fstring
from p2g import op
from p2g import symbol
from p2g import walkbase


class WalkExpr(walkbase.WalkBase):
    def _eval_ifs(self, its):
        for cond in its.ifs:
            if not self.visit(cond):
                return False
        return True

    def _enumerate_comps(self, iters):
        if iters:
            for el in self.visit(iters[0].iter):
                self.visit_store(iters[0].target, el)
                for _ in self._enumerate_comps(iters[1:]):
                    if self._eval_ifs(iters[0]):
                        yield
        else:
            yield

    def _visit_listcomp(self, node):
        with self.pushpop_funcns():
            return [self.visit(node.elt) for _ in self._enumerate_comps(node.generators)]

    def _visit_setcomp(self, node):
        with self.pushpop_funcns():
            return {self.visit(node.elt) for _ in self._enumerate_comps(node.generators)}

    def _visit_dictcomp(self, node):
        with self.pushpop_funcns():
            return {
                self.visit(node.key): self.visit(node.value)
                for _ in self._enumerate_comps(node.generators)
            }

    def _visit_ifexp(self, node):
        # turn x if tst else y
        # into   x * (tst ) + (y *  (1-tst))
        lhs = self.visit(node.body)
        rhs = self.visit(node.orelse)
        cond = self.visit(node.test)
        cond = cond != 0
        not_cond = cond == 0
        return lhs * cond + rhs * not_cond

    def _visit_compare(self, node):
        acc = self.visit(node.left)
        for cop, child in zip(node.ops, node.comparators):
            acc = op.make_vec_binop(op.a2opfo(type(cop)), acc, self.visit(child))
        return acc

    def _visit_boolop(self, node):
        if isinstance(node.op, ast.And):
            res = True
            for val in node.values:
                res = res and self.visit(val)
        else:
            res = False
            for val in node.values:
                res = res or self.visit(val)
        return res

    def _visit_joinedstr(self, node):
        return fstring.joinedstr(self, node)

    def _visit_binop(self, node):
        return op.make_vec_binop(
            op.a2opfo(type(node.op)),
            self.visit(node.left),
            self.visit(node.right),
        )

    def _visit_unaryop(self, node):
        return op.make_vec_binop(
            op.a2opfo(type(node.op)),
            self.visit(node.operand),
        )

    def _visit_subscript(self, node):
        obj = self.visit(node.value)
        idx = self.visit(node.slice)
        match node.ctx:
            case ast.Del():
                del obj[idx]
            case _:
                return obj[idx]
        return None

    def _visit_store_subscript(self, node, store_val):
        self.visit(node.value)[self.visit(node.slice)] = store_val

    def _visit_slice(self, node):
        return op.make_slice(
            self.visit(node.lower), self.visit(node.upper), self.visit(node.step)
        )

    def _visit_attribute(self, node):
        obj = self.visit(node.value)
        match node.ctx:
            case ast.Del():
                delattr(obj, node.attr)
            case _:
                res = getattr(obj, node.attr)
                symbol.Table.remember_load(node.attr, res)
                return res
        return None

    def _visit_store_attribute(self, node, store_val):
        symbol.Table.remember_store(node.attr, store_val)
        setattr(self.visit(node.value), node.attr, store_val)

    def _visit_name(self, node):
        return walkbase.handle_visit_name(self, node)

    def _visit_store_name(self, node, store_val):
        return walkbase.handle_visit_name_store(self, node, store_val)

    def _visit_dict(self, node):
        return {self.visit(p[0]): self.visit(p[1]) for p in zip(node.keys, node.values)}

    def _visit_set(self, node):
        return {self.visit(e) for e in node.elts}

    def _visit_list(self, node):
        return [self.visit(e) for e in node.elts]

    def _visit_tuple(self, node):
        return tuple(self.visit(e) for e in node.elts)

    def _visit_constant(self, node):
        return node.value



# flake8: noqa

import ast
import math
import typing

from p2g import err
from p2g import gbl
from p2g import nd
from p2g import scalar
from p2g import stat
from p2g import symbol
from p2g import vector


OptRes = typing.Optional[scalar.Scalar]
ScalarScalar = typing.Union[scalar.Scalar, int, float, bool]

op_byclass: typing.Dict[typing.Any, nd.Opinfo] = {}
allops: typing.Dict[str, nd.Opinfo] = {}


def get_constant(node):
    if isinstance(node, scalar.Constant):
        return node
    return None


def parensif(cond, thing):
    if cond:
        return "[" + thing + "]"
    return thing


def get_prec(thing):
    return getattr(thing, "prec", 20)


class Unop(scalar.Scalar):
    def __init__(self, opfo, child):
        super().__init__(opfo)
        self.child = child

    def get_address(self):
        return self.child

    def same(self, other):
        return self.opfo == other.opfo and gbl.same(self.child, other.child)

    def to_gcode(self, modifier=nd.NodeModifier.EMPTY):
        res = []
        if self.opfo.pyn == "#":
            try:
                addr = int(self.child)
                symbol.Table.add_varref(addr, gbl.iface.last_node)
            except TypeError:
                pass

        outer_prec = self.opfo.prec

        res.append(self.opfo.gname)
        # at least # is left right associative.
        inside_prec = get_prec(self.child)

        if self.opfo.pyn == "#":
            modifier |= nd.NodeModifier.ADDRESS

        res.append(
            parensif(
                outer_prec >= inside_prec,
                nd.to_gcode(self.child, modifier),
            )
        )

        return "".join(res)

    def __repr__(self):
        return f"[{self.opfo.pyn} {self.child}]"


def get_unop(node):
    if isinstance(node, Unop):
        return node
    return None


def get_nelements(node):
    if isinstance(node, vector.Vec):
        return node.nelements()

    return 1


def hashop(arg):
    return Unop(allops["#"], arg)


# make sure that whatever expression is in node
# ends up in a macro variable on


def reload(node: scalar.Scalar):
    unop = get_unop(node)
    if unop and unop.opfo.pyn == "#":
        if get_constant(unop.child):
            return node
    tmpaddr = gbl.iface.next_bss(1)
    tmpreg = hashop(tmpaddr)
    stat.append_set(tmpreg, node)
    return tmpreg


class Binop(scalar.Scalar):
    def __init__(self, opfo, lhs, rhs):
        super().__init__(opfo)
        assert opfo.pyn != "-rev"
        self.lhs = lhs
        self.rhs = rhs

    def same(self, other):
        return (
            self.opfo == other.opfo
            and gbl.same(self.lhs, other.lhs)
            and gbl.same(self.rhs, other.rhs)
        )

    def to_gcode(self, modi: nd.NodeModifier) -> str:

        if modi == nd.NodeModifier.ARGUMENT:

            return parensif(True, self.to_gcode(nd.NodeModifier.EMPTY))

        if self.opfo.g_func:
            res = [
                nd.to_gcode(self.lhs, modi),
                nd.to_gcode(self.rhs, modi),
            ]
            return f"{self.opfo.gname}[{','.join(res)}]"

        res = []

        outer_prec = self.opfo.prec

        def handle_term(node):
            return parensif(
                get_prec(node) < outer_prec,
                nd.to_gcode(node, modi),
            )

        res.append(handle_term(self.lhs))
        res.append(self.opfo.gname)
        res.append(handle_term(self.rhs))

        return " ".join(res)

    def __repr__(self):
        return f"({self.lhs}{self.opfo.pyn}{self.rhs})"


def make_scalar_binop(opfo, lhs, rhs):
    # can it be done right now?
    if isinstance(rhs, scalar.ConstantNone):
        return rhs
    if isinstance(lhs, scalar.ConstantNone):
        return lhs
    if isinstance(lhs, scalar.Constant) and isinstance(rhs, scalar.Constant):
        lval = lhs.value
        rval = rhs.value
        if isinstance(lval, float) or isinstance(rval, float):
            lval = float(lval)
            rval = float(rval)
        if opfo.pyn == "round":
            rval = int(rval)
        res = getattr(lval, opfo.mth)(rval)
        return scalar.Constant(res)
    # put contant on rhs if we can
    if (
        opfo.comm
        and isinstance(lhs, scalar.Constant)
        and not isinstance(rhs, scalar.Constant)
    ):
        lhs, rhs = rhs, lhs

    if opfo.opt and (res := opfo.opt(opfo, lhs, rhs)) is not None:
        return res

    return Binop(opfo, lhs, rhs)


def make_scalar_add(lhs, rhs):
    return make_scalar_binop(allops["+"], lhs, rhs)


def a2opfo(optyp):
    return op_byclass[optyp]


def opt_func(opfo, arg) -> OptRes:
    if not isinstance(arg, scalar.Constant):
        return None

    return scalar.wrap_scalar(opfo.lam(arg.to_float()))


def make_scalar_unop(opfo, child: ScalarScalar):
    child = scalar.wrap_scalar(child)
    if opfo.opt and (res := opfo.opt(opfo, child)) is not None:
        return res

    if isinstance(child, scalar.Constant):
        return scalar.wrap_scalar(getattr(child.value, opfo.mth)())

    return Unop(opfo, child)


def make_vec_binop(opfo, lhs, rhs=None, force_ourtype=False):

    # is doesn't have dunder method.
    if opfo.pyn == "is":
        return lhs is rhs

    if not force_ourtype and not isinstance(lhs, (int, float, nd.EBase)):
        return getattr(lhs, opfo.mth)(rhs)

    lhs = vector.wrap(lhs)

    if rhs is None:
        # actually unop.

        return vector.sorv_from_list(
            [make_scalar_unop(opfo, el) for el in lhs.everything()]
        )
    rhs = vector.wrap(rhs)
    return vector.sorv_from_list(
        [
            make_scalar_binop(opfo, lel, rel)
            for lel, rel in zip(lhs.everything(), rhs.forever())
        ]
    )


def make_scalar_func(fname, *args):
    ofo = allops[fname]
    return make_vec_binop(ofo, args[0], force_ourtype=True)


revop = {
    "+": "+",
    "-": "+",
    "*": "*",
    "/": "*",
}


# a  + k1  + k2 -> a + (k1+k2)
def opt_fold(opfo, lhs, rhs) -> OptRes:
    rop = revop.get(opfo.pyn)

    if (
        rop
        and isinstance(lhs, Binop)
        and lhs.opfo == opfo
        and lhs.rhs.is_constant
        and rhs.is_constant
    ):
        return make_scalar_binop(
            opfo,
            lhs.lhs,
            make_scalar_binop(
                allops[rop],
                lhs.rhs,
                rhs,
            ),
        )

    return None


def opt_mul(opfo, lhs, rhs) -> OptRes:
    match rhs.to_float():
        case 0.0:
            return scalar.Constant(0.0)
        case 1.0:
            return lhs
        case -1.0:
            return make_scalar_unop(allops["un-"], lhs)

    return opt_fold(opfo, lhs, rhs)


def opt_div(opfo, lhs, rhs) -> OptRes:
    match rhs.to_float():
        case None:
            return None
        case 0.0:
            err.compiler("Attempt to divide by zero.")
        case 1.0:
            return lhs
        case -1.0:
            return make_scalar_unop(allops["un-"], lhs)

    return opt_fold(opfo, lhs, rhs)


not_compop = {
    "==": "!=",
    "!=": "==",
    "<": ">=",
    ">": "<=",
    "<=": ">",
    ">=": "<",
}


def opt_not(_nd, arg) -> OptRes:
    if arg.is_constant:
        return scalar.Constant(not arg.value)

    # got not(comop, turn into (inverted compop))

    if notted := not_compop.get(arg.opfo.pyn):
        return make_scalar_binop(allops[notted], arg.lhs, arg.rhs)

    rhs = scalar.Constant(0)
    return make_scalar_binop(allops["!="], arg, rhs)


def opt_plus(_, arg) -> OptRes:
    return arg


# don't have binary not operator, make from xor.
def opt_invert(_, arg) -> OptRes:
    rhs = scalar.wrap_scalar(-1)
    return make_scalar_binop(allops["^"], arg, rhs)


# turn a1 > a2  == 0
# => (a1 <= a2)


def opt_eq(opfo, lhs: scalar.Scalar, rhs: scalar.Scalar) -> OptRes:
    if gbl.same(lhs, rhs):
        return scalar.Constant(opfo.pyn == "==")

    if not isinstance(rhs, scalar.Constant) or rhs.value != 0.0:
        return None

    if isinstance(lhs, Binop):
        if rev := not_compop.get(lhs.opfo.pyn):
            if opfo.pyn == "==":
                return make_scalar_binop(
                    allops[rev],
                    lhs.lhs,
                    lhs.rhs,
                )
            return lhs
    return None


def opt_matmul(_opfo, _lhs, _rhs) -> OptRes:  # for debug
    return None


def opt_add(opfo, lhs, rhs) -> OptRes:
    if (res := opt_fold(opfo, lhs, rhs)) is not None:
        return res
    if isinstance(rhs, Unop) and rhs.opfo.pyn == "un-":
        return make_scalar_binop(allops["-"], lhs, rhs.child)

    if isinstance(rhs, scalar.Constant):
        rhsv = rhs.to_float()
        if rhsv < 0.0:
            return make_scalar_binop(allops["-"], lhs, -rhsv)
        if rhsv == 0.0:
            return lhs

    return None


def opt_sub(opfo, lhs, rhs) -> OptRes:
    if (res := opt_fold(opfo, lhs, rhs)) is not None:
        return res

    # a -  - b -> a + b
    if isinstance(rhs, Unop) and rhs.opfo.pyn == "un-":
        return make_scalar_binop(allops["+"], lhs, rhs.child)

    if isinstance(rhs, scalar.Constant):
        rhsv = rhs.to_float()
        if rhsv == 0.0:
            return lhs

    return None


def unwrap_int(val):
    try:
        return val.to_int()
    except AttributeError:
        return val


def make_slice(low, high, step):
    return slice(
        unwrap_int(low),
        unwrap_int(high),
        unwrap_int(step),
    )


# breakpoint()
# vector.MemVec.make_hashop = make_hashop


def nd_install(opfo: nd.Opinfo):
    def make_multi_binop_tramp(lhs, rhs):
        return make_vec_binop(opfo, lhs, rhs)

    def make_multi_unop_tramp(child):
        return vector.sorv_from_list(
            [make_scalar_unop(opfo, cel) for cel in child.everything()]
        )

    def make_scalar_binop_tramp(lhs, rhs):
        return make_scalar_binop(opfo, lhs, scalar.wrap_scalar(rhs))

    def make_scalar_unop_tramp(child):
        return make_scalar_unop(opfo, child)

    if opfo.mth and isinstance(opfo.mth, str):
        if opfo.nargs == 1:
            setattr(vector.Vec, opfo.mth, make_multi_unop_tramp)
            setattr(scalar.Scalar, opfo.mth, make_scalar_unop_tramp)
        else:
            setattr(vector.Vec, opfo.mth, make_multi_binop_tramp)
            setattr(scalar.Scalar, opfo.mth, make_scalar_binop_tramp)


def reg(**kwargs):
    opi = nd.Opinfo(**kwargs)
    allops[opi.pyn] = opi
    op_byclass[opi.astc] = opi
    nd_install(opi)


def regfunc(name, lam):
    reg(
        pyn=name,
        lam=lam,
        gname=name.upper(),
        prec=20,
        nargs=1,
        opt=opt_func,
    )


regfunc("cos", lambda x: math.cos(math.radians(x)))
regfunc("sin", lambda x: math.sin(math.radians(x)))
regfunc("tan", lambda x: math.tan(math.radians(x)))
regfunc("acos", lambda x: math.degrees(math.acos(x)))
regfunc("asin", lambda x: math.degrees(math.asin(x)))
regfunc("atan", lambda x: math.degrees(math.atan(x)))
regfunc("exp", math.exp)
regfunc("ln", math.log)
regfunc("sqrt", math.sqrt)
regfunc("exists", id)
regfunc("fup", math.ceil)
regfunc("fix", math.floor)
regfunc("max", max)
regfunc("ground", round)


# fmt: off
reg(astc=ast.BitOr , pyn="|"  , mth="__or__" , gname="OR"    , prec=4, comm=True)
reg(astc=ast.BitXor, pyn="^"  , mth="__xor__", gname="XOR"   , prec=5, comm=True)
reg(astc=ast.BitAnd, pyn="&"  , mth="__and__", gname="AND"   , prec=6, comm=True)
reg(astc=ast.Lt    , pyn="<"  , mth="__lt__" , gname="LT"    , prec=10)
reg(astc=ast.LtE   , pyn="<=" , mth="__le__" , gname="LE"    , prec=10)
reg(astc=ast.NotEq , pyn="!=" , mth="__ne__" , gname="NE"    , prec=10, comm=True, opt=opt_eq)
reg(astc=ast.Eq    , pyn="==" , mth="__eq__" , gname="EQ"    , prec=10, comm=True, opt=opt_eq)
reg(astc=ast.Gt    , pyn=">"  , mth="__gt__" , gname="GT"    , prec=10)
reg(astc=ast.GtE   , pyn=">=" , mth="__ge__" , gname="GE"    , prec=10)
reg(astc=ast.Sub   , pyn="-"  , mth="__sub__", gname="-"     , prec=12, opt=opt_sub)
reg(astc=None      , pyn="-rev", mth="__rsub__", gname="-rev", prec=12)
reg(astc=ast.Add   , pyn="+"  , mth="__add__", gname="+"     , prec=12, comm=True, opt=opt_add)
reg(astc=None      , pyn="round" , mth="__round__", gname="round" , prec=12)
reg(astc=ast.Mult  , pyn="*"  , mth="__mul__" , gname="*"    , prec=13, comm=True, opt=opt_mul)
reg(astc=ast.Div   , pyn="/"  , mth="__truediv__", gname="/" , prec=13, opt=opt_div)
reg(astc=ast.FloorDiv , pyn="//"    , mth="__floordiv__" , gname="//"    , prec=13, opt=opt_div)
reg(astc=ast.Mod   , pyn="%"  , mth="__mod__" , gname="MOD"   , prec=14)
reg(astc=ast.Pow   , pyn="**" , mth="__pow__" , gname="POW"   , prec=15, g_func=True)
reg(astc=ast.USub  , pyn="un-", mth="__neg__" , gname="-"     , prec=14, nargs=1)

reg(astc=ast.UAdd  , pyn="un+" , mth="__pos__" , gname="+"     , prec=14, nargs=1, opt=opt_plus)
reg(astc=ast.MatMult, pyn="m*" , mth=""       , gname="ERRR"  , prec=20, nargs=2, opt=opt_matmul)
reg(astc=ast.Invert, pyn="un~" , mth="__invert__", gname="un~", prec=14, nargs=1, opt=opt_invert)
reg(astc=ast.Not   , pyn="unot", mth=""       , gname="unot"  , prec=14, nargs=1, opt=opt_not)
reg(astc=None      , pyn="abs" , mth="__abs__", gname="ABS"   , prec=15, nargs=1, g_func=True)
reg(astc=None      , pyn="#"   , mth=""       , gname="#"     , prec=19, nargs=1)
reg(astc=ast.Is    , pyn="is"  , mth=""       , gname="IS"    , prec=19, nargs=2)
# fmt:on
vector.local_hashop = lambda x, y: hashop(make_scalar_add(x, y))



from p2g import haas
from p2g import sys

from .axis import NAMES
from .builtin import abs  # pylint: disable=redefined-builtin
from .builtin import acos
from .gbl import Control
from .builtin import asin
from .builtin import atan
from .builtin import cos
from .builtin import exists
from .builtin import exp
from .builtin import fix
from .builtin import fup
from .builtin import ground
from .builtin import ln
from .builtin import sin
from .builtin import sqrt
from .builtin import tan
from .coords import Const
from .coords import CoType
from .coords import Fixed
from .coords import Var

from .goto import goto
from .scalar import Scalar
from .stat import code
from .stat import codenl
from .stat import com
from .stat import comment
from .symbol import Table
from .vector import TupleV
from .vector import Vec


__all__ = [
    "CoType",
    "Const",
    "Control",
    "Fixed",
    "NAMES",
    "Scalar",
    "Table",
    "TupleV",
    "VERSION",
    "Var",
    "Vec",
    "abs",
    "acos",
    "asin",
    "atan",
    "code",
    "codenl",
    "com",
    "comment",
    "cos",
    "exists",
    "exp",
    "fix",
    "fup",
    "goto",
    "ground",
    "haas",
    "sys",
    "ln",
    "sin",
    "sqrt",
    "tan",
]
VERSION = "0.3.1"



from p2g import axis
from p2g.coords import Fixed


# flake8: noqa

NAME = "HAAS NGC"
ORIGIN = [0, 0, 0]
NAXES = axis.Axes()
NO_SKIP = "M78"
MUST_SKIP = "M79"
FEED_FAST = 650.0
FEED_SLOW = 50.0
PROBE_FAST = 50.0
PROBE_SLOW = 10.0
PROBE_ON = ["G65 P9832"]
PROBE_H = Fixed[3](addr=2001)
OTS_ON = ["M59 P2", "G04 P1.0", "M59 P3"]
OTS_OFF = ["M69 P2", "M68 P3"]
SPINDLE_PROBE_ON = "G65 P9832"
SPINDLE_PROBE_OFF = "G65 P9833"
NO_LOOKAHEAD = ["G103 P1", "G04 P1", "G04 P1", "G04 P1"]
# MACHINE GEN BELOW
NULL                      = Fixed[    1](addr=   0)
MACRO_ARGUMENTS           = Fixed[   33](addr=   1)
GP_SAVED1                 = Fixed[   50](addr= 100)
PROBE_VALUES              = Fixed[   50](addr= 150)
GP_SAVED2                 = Fixed[   50](addr= 500)
PROBE_CALIB1              = Fixed[    6](addr= 550)
PROBE_R                   = Fixed[    3](addr= 556)
PROBE_CALIB2              = Fixed[   22](addr= 559)
GP_SAVED3                 = Fixed[  119](addr= 581)
GP_SAVED4                 = Fixed[  200](addr= 800)
INPUTS                    = Fixed[   64](addr=1000)
MAX_LOADS_XYZAB           = Fixed[    5](addr=1064)
RAW_ANALOG                = Fixed[   10](addr=1080)
FILTERED_ANALOG           = Fixed[    8](addr=1090)
SPINDLE_LOAD              = Fixed[    1](addr=1098)
MAX_LOADS_CTUVW           = Fixed[    5](addr=1264)
TOOL_TBL_FLUTES           = Fixed[  200](addr=1601)
TOOL_TBL_VIBRATION        = Fixed[  200](addr=1801)
TOOL_TBL_OFFSETS          = Fixed[  200](addr=2001)
TOOL_TBL_WEAR             = Fixed[  200](addr=2201)
TOOL_TBL_DROFFSET         = Fixed[  200](addr=2401)
TOOL_TBL_DRWEAR           = Fixed[  200](addr=2601)
ALARM                     = Fixed[    1](addr=3000)
T_MS                      = Fixed[    1](addr=3001)
T_HR                      = Fixed[    1](addr=3002)
SINGLE_BLOCK_OFF          = Fixed[    1](addr=3003)
FEED_HOLD_OFF             = Fixed[    1](addr=3004)
MESSAGE                   = Fixed[    1](addr=3006)
YEAR_MONTH_DAY            = Fixed[    1](addr=3011)
HOUR_MINUTE_SECOND        = Fixed[    1](addr=3012)
POWER_ON_TIME             = Fixed[    1](addr=3020)
CYCLE_START_TIME          = Fixed[    1](addr=3021)
FEED_TIMER                = Fixed[    1](addr=3022)
CUR_PART_TIMER            = Fixed[    1](addr=3023)
LAST_COMPLETE_PART_TIMER  = Fixed[    1](addr=3024)
LAST_PART_TIMER           = Fixed[    1](addr=3025)
TOOL_IN_SPIDLE            = Fixed[    1](addr=3026)
SPINDLE_RPM               = Fixed[    1](addr=3027)
PALLET_LOADED             = Fixed[    1](addr=3028)
SINGLE_BLOCK              = Fixed[    1](addr=3030)
AGAP                      = Fixed[    1](addr=3031)
BLOCK_DELETE              = Fixed[    1](addr=3032)
OPT_STOP                  = Fixed[    1](addr=3033)
TIMER_CELL_SAFE           = Fixed[    1](addr=3196)
TOOL_TBL_DIAMETER         = Fixed[  200](addr=3201)
TOOL_TBL_COOLANT_POSITION = Fixed[  200](addr=3401)
M30_COUNT1                = Fixed[    1](addr=3901)
M30_COUNT2                = Fixed[    1](addr=3902)
PREV_BLOCK                = Fixed[   13](addr=4001)
PREV_WCS                  = Fixed[    1](addr=4014)
PREV_BLOCK_ADDRESS        = Fixed[   26](addr=4101)
LAST_TARGET_POS           = Fixed[NAXES](addr=5001)
MACHINE_POS               = Fixed[NAXES](addr=5021)
WORK_POS                  = Fixed[NAXES](addr=5041)
SKIP_POS                  = Fixed[NAXES](addr=5061)
TOOL_OFFSET               = Fixed[NAXES](addr=5081)
G52                       = Fixed[NAXES](addr=5201)
G54                       = Fixed[NAXES](addr=5221)
G55                       = Fixed[NAXES](addr=5241)
G56                       = Fixed[NAXES](addr=5261)
G57                       = Fixed[NAXES](addr=5281)
G58                       = Fixed[NAXES](addr=5301)
G59                       = Fixed[NAXES](addr=5321)
TOOL_TBL_FEED_TIMERS      = Fixed[  100](addr=5401)
TOOL_TBL_TOTAL_TIMERS     = Fixed[  100](addr=5501)
TOOL_TBL_LIFE_LIMITS      = Fixed[  100](addr=5601)
TOOL_TBL_LIFE_COUNTERS    = Fixed[  100](addr=5701)
TOOL_TBL_LIFE_MAX_LOADS   = Fixed[  100](addr=5801)
TOOL_TBL_LIFE_LOAD_LIMITS = Fixed[  100](addr=5901)
NGC_CF                    = Fixed[    1](addr=6198)
G154_P1                   = Fixed[NAXES](addr=7001)
G154_P2                   = Fixed[NAXES](addr=7021)
G154_P3                   = Fixed[NAXES](addr=7041)
G154_P4                   = Fixed[NAXES](addr=7061)
G154_P5                   = Fixed[NAXES](addr=7081)
G154_P6                   = Fixed[NAXES](addr=7101)
G154_P7                   = Fixed[NAXES](addr=7121)
G154_P8                   = Fixed[NAXES](addr=7141)
G154_P9                   = Fixed[NAXES](addr=7161)
G154_P10                  = Fixed[NAXES](addr=7181)
G154_P11                  = Fixed[NAXES](addr=7201)
G154_P12                  = Fixed[NAXES](addr=7221)
G154_P13                  = Fixed[NAXES](addr=7241)
G154_P14                  = Fixed[NAXES](addr=7261)
G154_P15                  = Fixed[NAXES](addr=7281)
G154_P16                  = Fixed[NAXES](addr=7301)
G154_P17                  = Fixed[NAXES](addr=7321)
G154_P18                  = Fixed[NAXES](addr=7341)
G154_P19                  = Fixed[NAXES](addr=7361)
G154_P20                  = Fixed[NAXES](addr=7381)
ATM_ID                    = Fixed[    1](addr=8500)
ATM_PERCENT               = Fixed[    1](addr=8501)
ATM_TOTAL_AVL_USAGE       = Fixed[    1](addr=8502)
ATM_TOTAL_AVL_HOLE_COUNT  = Fixed[    1](addr=8503)
ATM_TOTAL_AVL_FEED_TIME   = Fixed[    1](addr=8504)
ATM_TOTAL_AVL_TOTAL_TIME  = Fixed[    1](addr=8505)
ATM_NEXT_TOOL_NUMBER      = Fixed[    1](addr=8510)
ATM_NEXT_TOOL_LIFE        = Fixed[    1](addr=8511)
ATM_NEXT_TOOL_AVL_USAGE   = Fixed[    1](addr=8512)
ATM_NEXT_TOOL_HOLE_COUNT  = Fixed[    1](addr=8513)
ATM_NEXT_TOOL_FEED_TIME   = Fixed[    1](addr=8514)
ATM_NEXT_TOOL_TOTAL_TIME  = Fixed[    1](addr=8515)
TOOL_ID                   = Fixed[    1](addr=8550)
TOOL_FLUTES               = Fixed[    1](addr=8551)
TOOL_MAX_VIBRATION        = Fixed[    1](addr=8552)
TOOL_LENGTH_OFFSETS       = Fixed[    1](addr=8553)
TOOL_LENGTH_WEAR          = Fixed[    1](addr=8554)
TOOL_DIAMETER_OFFSETS     = Fixed[    1](addr=8555)
TOOL_DIAMETER_WEAR        = Fixed[    1](addr=8556)
TOOL_ACTUAL_DIAMETER      = Fixed[    1](addr=8557)
TOOL_COOLANT_POSITION     = Fixed[    1](addr=8558)
TOOL_FEED_TIMER           = Fixed[    1](addr=8559)
TOOL_TOTAL_TIMER          = Fixed[    1](addr=8560)
TOOL_LIFE_LIMIT           = Fixed[    1](addr=8561)
TOOL_LIFE_COUNTER         = Fixed[    1](addr=8562)
TOOL_LIFE_MAX_LOAD        = Fixed[    1](addr=8563)
TOOL_LIFE_LOAD_LIMIT      = Fixed[    1](addr=8564)
THERMAL_COMP_ACC          = Fixed[    1](addr=9000)
THERMAL_SPINDLE_COMP_ACC  = Fixed[    1](addr=9016)
GVARIABLES3               = Fixed[ 1000](addr=10000)
PROBE_VALUES_             = Fixed[   50](addr=10150)
GPS1                      = Fixed[  200](addr=10200)
GPNS1                     = Fixed[  100](addr=10400)
GPS2                      = Fixed[   50](addr=10500)
PROBE_CALIB_              = Fixed[   50](addr=10550)
GPS3                      = Fixed[  100](addr=10600)
GPNS3                     = Fixed[  100](addr=10700)
GPS4                      = Fixed[  200](addr=10800)
INPUTS1                   = Fixed[  256](addr=11000)
OUTPUT1                   = Fixed[  256](addr=12000)
FILTERED_ANALOG1          = Fixed[   13](addr=13000)
COOLANT_LEVEL             = Fixed[    1](addr=13013)
FILTERED_ANALOG2          = Fixed[   50](addr=13014)
SETTING                   = Fixed[10000](addr=20000)
PARAMETER                 = Fixed[10000](addr=30000)
TOOL_TYP                  = Fixed[  200](addr=50001)
TOOL_MATERIAL             = Fixed[  200](addr=50201)
CURRENT_OFFSET            = Fixed[  200](addr=50601)
CURRENT_OFFSET2           = Fixed[  200](addr=50801)
VPS_TEMPLATE_OFFSET       = Fixed[  100](addr=51301)
WORK_MATERIAL             = Fixed[  200](addr=51401)
VPS_FEEDRATE              = Fixed[  200](addr=51601)
APPROX_LENGTH             = Fixed[  200](addr=51801)
APPROX_DIAMETER           = Fixed[  200](addr=52001)
EDGE_MEASURE_HEIGHT       = Fixed[  200](addr=52201)
TOOL_TOLERANCE            = Fixed[  200](addr=52401)
PROBE_TYPE                = Fixed[  200](addr=52601)
PROBE = SKIP_POS
WORK = WORK_POS
MACHINE = MACHINE_POS
G53 = MACHINE_POS
# MACHINE GEN ABOVE



import ast
import builtins
import contextlib
import dataclasses
import enum
import typing

from typing import Any

from p2g import err
from p2g import gbl
from p2g import nd
from p2g import stat
from p2g import symbol


UNDEF = gbl.sentinel()
GLOBAL = gbl.sentinel()
NONLOCAL = gbl.sentinel()


class NS(enum.IntEnum):
    MODULE = enum.auto()
    FUNCTION = enum.auto()
    CLASS = enum.auto()


@dataclasses.dataclass
class Namespace:
    parent: "Namespace"
    guts: typing.Dict
    nstype: NS

    def __init__(self, nstype=NS.MODULE):
        self.nstype = nstype
        self.guts = {}
        self.parent = typing.cast(Namespace, None)

    @property
    def is_function(self):
        return self.nstype == NS.FUNCTION

    @property
    def is_class(self):
        return self.nstype == NS.CLASS

    @property
    def is_module(self):
        return self.nstype == NS.MODULE

    def get(self, key):
        return self.guts.get(key, UNDEF)

    def __getitem__(self, key):
        return self.guts[key]

    def __setitem__(self, key, value):
        self.guts[key] = value

    def __delitem__(self, key):
        del self.guts[key]

    def __contains__(self, key):
        return key in self.guts


def scan_namespaces(ns):
    yield ns
    ns = ns.parent
    while ns:
        if not ns.is_class:
            yield ns
        ns = ns.parent


def resolve_nonlocal(name, ns):
    while ns:
        res = ns.get(name)
        if res is not UNDEF and res is not NONLOCAL:
            if ns.is_module:
                break
            return ns
        ns = ns.parent
    raise err.CompilerError(f"No binding for nonlocal '{name}'.")


def lookup_to_ns(res, self, nid):
    if res is GLOBAL:
        dstns = self.module_ns
    elif res is NONLOCAL:
        dstns = resolve_nonlocal(nid, self.ns.parent)
    else:
        dstns = self.ns
    return dstns


def find_ns(first_ns, nid):
    for ns in scan_namespaces(first_ns):
        res = ns.get(nid)
        if res is UNDEF:
            continue
        return res, ns
    return UNDEF, None


def _handle_visit_name_load(self, node):
    res, ns = find_ns(self.ns, node.id)
    if ns is None:
        try:
            return getattr(builtins, node.id)
        except AttributeError as exn:
            raise err.CompilerError(f"Name '{node.id}' is not defined.") from exn
    if res is GLOBAL:
        res = self.module_ns.get(node.id)
    elif res is NONLOCAL:
        ns = resolve_nonlocal(node.id, ns.parent)
        res = ns[node.id]
    assert res is not UNDEF
    symbol.Table.remember_load(node.id, res)
    return res


def handle_visit_name_del(self, node):
    res = self.ns.get(node.id)
    if res is UNDEF:
        err.compiler(f"Name '{node.id}' is not defined.")
    dstns = lookup_to_ns(res, self, node.id)
    del dstns[node.id]


def handle_visit_name_store(self, node, store_val):
    res = self.ns.get(node.id)
    if isinstance(res, nd.EBase):
        # variable is already in use, instead of
        # overwriting it,  treat as if writing to
        # <foo>.var
        #        breakpoint()
        setattr(res, "var", store_val)
        return
    dstns = lookup_to_ns(res, self, node.id)
    symbol.Table.remember_store(node.id, store_val)
    dstns[node.id] = store_val


def handle_visit_name(self, node):
    match node.ctx:
        case ast.Del():
            handle_visit_name_del(self, node)
        # in augmented assign, name arrives as with load ctx, but
        # must also be stored there, so catch both.
        case _:
            return _handle_visit_name_load(self, node)
    return None


@dataclasses.dataclass
class LoopContext:
    lcontinue: stat.Label
    lbreak: stat.Label
    lorelse: stat.Label
    prev: "LoopContext"

    def __init__(self, prev):
        self.prev = prev
        self.lcontinue = stat.next_label()
        self.lbreak = stat.next_label()
        self.lorelse = stat.next_label()


class WalkBase:
    ns: Namespace
    module_ns: Namespace
    file_name: str
    func_name: str

    def __init__(self):
        self.file_name = ""
        self.func_name = ""
        self.module_ns = Namespace()
        self.ns = self.module_ns
        self.loop = typing.cast(LoopContext, None)
        self.module_ns = typing.cast(Namespace, None)

    # def call(self, *_args, **_kwargs):
    #     raise NotImplementedError
    def visit_slist(self, lst):
        res = None
        for el in lst:
            res = self.visit(el)
        return res

    def update_lastplace(self, node):
        gbl.iface.last_node = node
        node.file_name = self.file_name

    def visit_fail(self, node, *_):
        name = node.__class__.__name__
        err.compiler(f"Feature '{name.lower()}' not implemented.")

    def _visit_return(self, node):
        if not self.ns.is_function:
            err.compiler("'return' outside function.")
        return self.visit(node.value)

    def visit(self, node) -> Any:
        if node is None:
            return None
        self.update_lastplace(node)
        method = "_visit_" + node.__class__.__name__.lower()
        visitor: typing.Callable[[ast.AST], Any] = getattr(self, method, self.visit_fail)

        return visitor(node)

    def visit_store(self, node, store_val) -> None:
        self.update_lastplace(node)
        method = "_visit_store_" + node.__class__.__name__.lower()
        visitor: typing.Callable[[ast.AST, ast.AST], None] = getattr(
            self, method, self.visit_fail
        )
        visitor(node, store_val)

    def _visit_expr(self, node):
        self.visit(node.value)

    @contextlib.contextmanager
    def pushpopns(self, ns):
        prev = self.ns
        self.ns = ns
        yield ns
        self.ns = prev

    def pushpop_with_parent(self, ns):
        ns.parent = self.ns
        return self.pushpopns(ns)

    def pushpop_funcns(self):
        return self.pushpop_with_parent(Namespace(NS.FUNCTION))

    def pushpop_classns(self):
        return self.pushpop_with_parent(Namespace(NS.CLASS))


class WalkNS(WalkBase):
    def _visit_global(self, node):
        for name in node.names:
            if name in self.ns and self.ns[name] is not GLOBAL:  # hard to test right
                err.compiler(f"Name '{name}' before global.")
            # Don't store GLOBAL in the top-level namespace
            if self.ns.parent:
                self.ns[name] = GLOBAL

    def _visit_nonlocal(self, node):
        if self.ns.is_module:  # hard to test right
            err.compiler("'nonlocal' declaration not allowed at module level.")
        for name in node.names:
            self.ns[name] = NONLOCAL



from p2g import nd


class Scalar(nd.EBase):
    opfo: nd.Opinfo
    is_none_constant = False
    is_constant = False

    def __init__(self, opfo):
        self.opfo = opfo

    @property
    def prec(self):
        return self.opfo.prec

    def forever(self):
        while True:
            yield self

    def to_float(self):
        return None

    def everything(self):
        yield self

    # placeholder to east typechecking,
    # overwitten by op install machines.
    def __add__(self, _other):  # no cover
        pass

    def __lt__(self, _other):  # no cover
        pass

    def __div__(self, _other):  # no cover
        pass

    def __int__(self) -> int:
        raise TypeError


class ConstantBase(Scalar):
    is_constant = True


class Constant(ConstantBase):
    _value: int | float

    def __init__(self, value: int | float):
        super().__init__(nd.const_nd)
        self._value = value

    # def get_address(self):
    #     err.compiler("Can't take address of constant.")
    @property
    def value(self):
        return self._value

    def same(self, other: "Constant"):
        return isinstance(other, Constant) and other.value == self._value

    def to_int(self) -> int:
        return int(self)

    def to_float(self):
        return float(self._value)

    def to_gcode(self, modifier=nd.NodeModifier.EMPTY) -> str:
        return nd.to_gcode_from_float(self._value, modifier)

    def __bool__(self):
        if self._value:
            return True
        return False

    def __int__(self) -> int:
        return int(self._value)

    def __repr__(self):
        return f"<{self._value}>"

    def __str__(self):
        return str(self._value)


class ConstantNone(ConstantBase):
    is_none_constant = True

    def __init__(self):
        super().__init__(nd.const_nd)

    def to_gcode(self, _modifier) -> str:
        return "<none>"


class ConstantStr(ConstantBase):
    value: str

    def __init__(self, value: str):
        super().__init__(nd.const_nd)
        self.value = value

    def to_gcode(self, modifier=nd.NodeModifier.EMPTY) -> str:
        # as an integer literal or as a string.
        if modifier & nd.NodeModifier.ARGUMENT:
            return f"{ord(self.value)}."
        return f"'{self.value}'"


def wrap_scalar(thing) -> Scalar:
    if isinstance(thing, str):
        return ConstantStr(thing)
    if thing is None:
        return ConstantNone()
    if isinstance(thing, (int, float, bool)):
        return Constant(thing)
    if isinstance(thing, Scalar):
        return thing
    return thing.to_scalar()


# make a range between wrapped items and yield ints
def urange(start, stop, step=1):
    stop = int(stop)
    step = int(step)
    tmp = list(range(int(start), stop, step))
    for el in tmp:
        yield el



import abc
import collections
import enum
import typing

from p2g import gbl


class Group(enum.IntEnum):
    IGNORE = enum.auto()
    USER = enum.auto()
    VECTOR = enum.auto()
    GOTO = enum.auto()


class Entry(typing.NamedTuple):
    major_key: Group
    key: str
    info: str

    def sort_key(self):
        return f"{self.major_key}{self.key}{self.info}"


# marks class as being able to emit special symtab table entries.
# pylint: disable=too-few-public-methods
class HasToSymTabEntry(abc.ABC):
    @abc.abstractmethod
    def to_symtab_entry(self, _addrs_used) -> typing.Tuple[Group, str]:
        return Group.IGNORE, ""


# symbol tables made by watching every assignment from a vector
# constructor to a name and remembering all the address constants
# used.  walk through all address constants, find the symbols
# assocaited and mark them as useful.  then print them out.
class Table:
    print_all = False
    name_to_thing: typing.Dict[str, list[HasToSymTabEntry]] = collections.defaultdict(
        list
    )
    addrs_used: typing.Set[int] = set()

    @classmethod
    def _remember(cls, key, thing):
        if key in ("xyz", "xy", "x", "y", "z", "val"):
            return
        if cls.print_all or isinstance(thing, HasToSymTabEntry):
            cls.name_to_thing[key].append(thing)

    @classmethod
    def remember_load(cls, key, thing):
        cls._remember(key, thing)

    @classmethod
    def remember_store(cls, key, thing):
        cls._remember(key, thing)

    @classmethod
    # remember a used macro var.
    def add_varref(cls, addr, _pos):
        cls.addrs_used.add(int(addr))

    @classmethod
    def yield_table(cls):
        if not gbl.Control.symbol_table and not cls.print_all:
            return
        lsize = 0
        rsize = 0

        def entries():
            nonlocal lsize
            nonlocal rsize
            for key, things in cls.name_to_thing.items():
                lsize = max(lsize, len(key))
                for thing in things:
                    major_key, info = thing.to_symtab_entry(cls.addrs_used)
                    rsize = max(rsize, len(info))
                    if key and info:
                        yield Entry(major_key, key, info)

        prev = Entry(Group.IGNORE, "", "")
        yield "( Symbol Table )"
        for entry in sorted(entries(), key=Entry.sort_key):
            if prev == entry:
                continue
            if prev.major_key != entry.major_key:
                yield ""
            yield " ( " + entry.key.ljust(lsize) + " : " + entry.info.ljust(rsize) + " )"
            prev = entry
        # blank line after symbol table
        yield ""

    @classmethod
    def reset(cls):
        cls.name_to_thing = collections.defaultdict(list)
        cls.addrs_used = set()
        gbl.Control.symbol_table = False


class TSUB:
    pass
